# -*- coding: utf-8 -*-
"""
Example analysis of Scanette logs - new CSV format.

Reads Scanette CSV files with these columns:
# Columns Docs from Frederick, 2019-10-18.
  0: id is an identifier of the line (some numbers may be missing)
  1: timestamp is in Linux format with three extra digits for milliseconds.
  2: sessionID provides the identifier of the session - each client is different.
  3: objectInstance is the object instance on which the operation is invoked.
  4: operation is the name of the operation (action).
  5: parameters is a list of the parameter values, or [] if there are no parameters.
  6: result is the status code returned (? means that the operation does
    not return anything - void)

Created on Thu Oct 17 16:45:39 2019

@author: utting@usc.edu.au
"""

import csv
from pathlib import Path
from datetime import datetime, date, time
from sklearn.cluster import MeanShift
from sklearn.preprocessing import MinMaxScaler
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
import matplotlib.cm as pltcm
import matplotlib.markers as pltmarkers
# This import registers the 3D projection, but is otherwise unused.
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import
from collections import Counter
import sys

import agilkia

# %%

DISPLAY_TIME = 3
TEST_MODE = "--test" in sys.argv
if TEST_MODE:
    print(f"Running in test mode - all figures will close after {DISPLAY_TIME} seconds.")

# %%

def read_traces_csv(path: Path) -> agilkia.TraceSet:
    # print("now=", datetime.now().timestamp())
    with path.open("r") as input:
        trace1 = agilkia.Trace([])
        for line in csv.reader(input):
            # we ignore the line id.
            timestr = line[1].strip()
            timestamp = date.fromtimestamp(int(timestr) / 1000.0)
            # print(timestr, timestamp.isoformat())
            sessionID = line[2].strip()
            objInstance = line[3].strip()
            action = line[4].strip()
            paramstr = line[5].strip()
            result = line[6].strip()
            # now we identify the main action, inputs, outputs, etc.
            if paramstr == "[]":
                inputs = {}
            else:
                if  paramstr.startswith("[") and paramstr.endswith("]"):
                    paramstr = paramstr[1:-1]
                inputs = {"param" : paramstr}
            if result == "?":
                outputs = {}
            else:
                outputs = {'Status': float(result)}
            others = {
                    'timestamp': timestamp,
                    'sessionID': sessionID,
                    'object': objInstance
                    }
            event = agilkia.Event(action, inputs, outputs, others)
            trace1.append(event)
    traceset = agilkia.TraceSet([])
    traceset.append(trace1)
    return traceset


# %% Read traces and save in the Agilkia JSON format.

french_chars = {'scanner': '.', 'abandon': 'a', 'supprimer': 's',
                'ajouter': '+', 'debloquer': 'd', 'fermerSession': 'f',
                'ouvrirSession': 'o', 'payer': 'p', 'transmission': 't'}
english_chars = {'scanner': '.',        # 'scan'
                 'abandon': 'a',        # 'abandon'
                 'supprimer': 'd',      # 'delete'
                 'ajouter': '+',        # 'add'
                 'debloquer': 'u',      # 'unlock'
                 'fermerSession': 'c',  # 'closeSession'
                 'ouvrirSession': 'o',  # 'openSession'
                 'payer': 'p',          # 'pay'
                 'transmission': 't'    # 'transmit'
                 }
traceset = read_traces_csv(Path("127.0.0.1-1571403244552.csv"))
traceset.set_event_chars(english_chars)
    # was: {"scanner": ".", "abandon": "a", "supprimer": "s", "ajouter": "+"}

print("Event chars:\n  ", traceset.get_event_chars())  # default summary char for each kind of event.

print(f"One long trace, saved to log_one.json.  Length={len(traceset[0])} events.")
traceset.save_to_json(Path("log_one.json"))
print(str(traceset[0])[:200], "etc")  # everything is in one big trace initially.


# %% Split into separate traces, first based on Scanette number.

print("\n\n==== grouped by sessionID number ====")
traceset2 = traceset.with_traces_grouped_by("sessionID", property=True)
for tr in traceset2[0:10]:
    print("   ", tr)
print(f"    etc")
print(f"{len(traceset2)} traces now, saved into log_split.json")
traceset2.save_to_json(Path("log_split.json"))

# %% Sometimes one needs to split into smaller traces each time an event happens.

#print("\n\n==== split before each S.debloquer ====")
# traceset3 = traceset2.with_traces_split(start_action="S.debloquer")
traceset3 = traceset2

# %% Get some data about each trace, to use for clustering traces.

data = traceset3.get_trace_data(method="action_counts")  # or add: method="action_status_counts"
print(data.sum().sort_values())

# %% Now cluster the traces (default MeanShift)

clusterer = MeanShift()
normalizer = MinMaxScaler()
num_clusters = traceset3.create_clusters(data, algorithm=clusterer, normalizer=normalizer)
print(num_clusters, "clusters found - see clusters_out.txt")
with Path("clusters_out.txt").open("w") as out:
    for i in range(num_clusters):
        out.write(f"Cluster {i}:\n")
        for tr in traceset3.get_cluster(i):
            out.write(f"    {tr}\n")

# %% Count number of traces in each cluster

counts = Counter(traceset3.get_clusters())
count_pairs = sorted(list(counts.items()))
print("cluster sizes:")
for c,n in count_pairs:
    print(f"    {c:3d}  {n:4d}")

# or graph them:
# plt.bar([x for (x,y) in count_pairs], [y for (x,y) in count_pairs], log=True)

# %% Visualise clusters (using default TSNE) - not great for this example

traceset3.visualize_clusters(block=(not TEST_MODE))

# %% Visualise cluster using PCA to map them into 2D.

# Here is a reasonably good choice of markers for the 13 Scanner clusters.
# (good choice depends upon how many clusters, how many points in each cluster, overlap, etc.)
markers = "1234+x*.<^>vo" # "sphPXd"
xlim = (-1.0,+3.0)
ylim = (-0.65, +0.9)
vis = PCA(n_components=2)
traceset3.visualize_clusters(algorithm=vis, xlim=xlim, ylim=ylim,
                             markers=markers,
                             markersize=9,
                             filename="scanette_clusters.pdf",
                             block=(not TEST_MODE))

# %% Print PCA dimensions

# print(traceset3.get_event_chars())
print(data.columns)
print("PCA components:")
print(vis.components_)
print("PCA explained variance ratios:")
print(vis.explained_variance_ratio_)
print("Total explained variance ratio:", sum(vis.explained_variance_ratio_))


# %% Read traces from functional tests

tests = read_traces_csv(Path("tests_1571927354131.csv"))
event_chars3 = traceset3.get_event_chars()
tests.set_event_chars(event_chars3)
print(f"Reading system tests - one trace, {len(tests[0])} events.")

# %% Split tests into separate traces

tests2 = tests.with_traces_grouped_by("sessionID", property=True)
assert tests2.get_event_chars() == event_chars3

print(f"Split into {len(tests2)} test traces, saved into tests_split.json")
tests2.save_to_json(Path("tests_split.json"))

# %% Do some statistical analysis about action frequencies

test_df = tests2.to_pandas()
print(test_df)
print("\nAction frequencies:\n", test_df["Action"].value_counts())
print("\nStatus frequencies:\n", test_df["Status"].value_counts())

# %% Get the action-count data for the test traces.

test_data = tests2.get_trace_data(method="action_counts")
# NOTE: we must make sure this has same columns in same order!
# print("data.columns     ", list(data.columns))
# print("test_data.columns", list(test_data.columns))

# add the missing 'ajouter' column (all zeroes) before 'fermerSession' (5)
if len(data.columns) < 10:
    test_data.insert(1, 'ajouter', 0)
else:
    test_data.insert(0, 'none', 0)
    keep = ['debloquer_0', 'scanner_0', 'scanner_-2', 'transmission_0', 'abandon_0',
       'ouvrirSession_0', 'none', 'fermerSession_0',
       'payer_-102', 'payer_-42', 'payer_0', 'payer_-4',
       'none',  # transmission_1
       'payer_-1', 'payer_2',
       'none', 'none', 'none', 'none',  # extras
       'supprimer_-2', 'scanner_-1']
    test_data = test_data[keep]
# check that names of columns are the same (but status values can be different)
if not all([c1[:5]==c2[:5] for (c1,c2) in zip(data.columns,test_data.columns) if c2 != "none"]):
    print("ERROR: test columns are not the same as data.columns.")
    print("data.columns=", data.columns)
    print("test.columns=", test_data.columns)

# %% Put the test traces into the original clusters from above.

# NOTE: with action_status_counts, we cannot use same normalizer here, because of error:
#
#  File "/Users/utting/anaconda3/envs/bus-testing/lib/python3.7/site-packages/sklearn/preprocessing/data.py", line 389, in transform
#    X *= self.scale_
#
#ValueError: operands could not be broadcast together with shapes (30,20) (21,) (30,20)

test_clusters = clusterer.predict(normalizer.transform(test_data))
print("tests clustered - see test_clusters_out.txt")
print(test_clusters)
with Path("test_clusters_out.txt").open("w") as out:
    for i in range(num_clusters):
        out.write(f"Cluster {i}:\n")
        for pos, cluster in enumerate(test_clusters):
            if cluster == i:
                out.write(f"    {tests2[pos]}\n")

# %% Check that the tests are using the same event_chars for display.

print(tests2.get_event_chars())

# %% Count the tests in each cluster and graph them vs customer clusters

test_counts = Counter(test_clusters)
test_pairs = sorted(list(test_counts.items()))
plt.bar([x for (x,y) in count_pairs], [y for (x,y) in count_pairs], width=0.4, log=True)
plt.bar([x+0.4 for (x,y) in test_pairs], [y for (x,y) in test_pairs], width=0.4, log=True)
plt.savefig("cust_vs_tests_log.pdf")
plt.show(block=(not TEST_MODE))
print("Number of tests =", sum([v for (_,v) in test_pairs]))

# %% check test cluster results (when regression testing)

assert test_pairs == [(0, 7), (1, 9), (2, 2), (6, 2), (7, 8), (10, 1), (11, 1)]

# %% Visualise them into the SAME clusters by passing in the pre-fitted models

#print(vis.transform(test_data))
n = tests2.create_clusters(test_data, algorithm=clusterer, normalizer=normalizer, fit=False)
print("clusters=", n)
print(tests2.get_clusters())
print(len(tests2.get_clusters()))

# %%

tests2.visualize_clusters(algorithm=vis, fit=False,
                          xlim=xlim, ylim=ylim,
                          markers=markers, markersize=9,
                          filename="test_clusters.pdf",
                          block=(not TEST_MODE))


# %%

print("Trying 3D plot...")
vis3 = PCA(n_components=3)
traceset3.visualize_clusters(algorithm=vis3, block=(not TEST_MODE))

plot_data = vis3.transform(normalizer.transform(data))
print(plot_data[:10])

plot_test = vis3.transform(normalizer.transform(test_data))
print(plot_test[:10])

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
colors = pltcm.get_cmap('hsv')

ax.scatter(plot_data[:, 0], plot_data[:, 1], plot_data[:, 2],
           c=traceset3.get_clusters(), cmap=colors)
ax.scatter(plot_test[:, 0], plot_test[:, 1], plot_test[:, 2],
           c=tests.get_clusters(), cmap=colors, marker="+")
plt.show(block=(not TEST_MODE))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Example analysis of Scanette logs to generate missing tests.

@author: utting@usc.edu.au
"""

from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import AdaBoostClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import LinearSVC
from sklearn.dummy import DummyClassifier

from sklearn.preprocessing import MinMaxScaler
from sklearn.decomposition import PCA
from sklearn.model_selection import cross_val_score
from sklearn.pipeline import Pipeline
from sklearn.tree import export_graphviz, export_text
import matplotlib.pyplot as plt
from pathlib import Path
import random
import timeit
import sys

import agilkia

# %%

DISPLAY_TIME = 3
TEST_MODE = "--test" in sys.argv
if TEST_MODE:
    print(f"Running in test mode - all figures will close after {DISPLAY_TIME} seconds.")

# %%

INPUT = Path("log_split.json")
cust = agilkia.TraceSet.load_from_json(INPUT)
print(f"Loaded {len(cust)} customer traces from {INPUT}")
print(cust.get_event_chars())
cust.meta_data["dataset"] = "Scanette customer traces"


# %% Get some data about each trace, to use for clustering traces.

data = cust.get_trace_data(method="action_counts")  # or add: method="action_status_counts"
print(data.sum().sort_values())

# %% Now cluster the traces (default MeanShift)

num_clusters = cust.create_clusters(data)
print(num_clusters, "clusters found")
for i in range(num_clusters):
    print(f"Cluster {i}:")
    n = 0
    for tr in cust.get_cluster(i):
        n += 1
        if n > 5:
            print("    ...")
            break
        print(f"    {tr}")


# %% Visualise clusters (using TSNE)

vis = PCA(n_components=2)
cust.visualize_clusters(algorithm=vis, block=(not TEST_MODE))
if TEST_MODE:
    plt.pause(3)
    plt.close()

# %%

print(cust.meta_data)

# %% The event signatures for Scanette

signature = {
    "abandon": {"input":{}, "output":{"Status":"int"}},
    "ajouter": {"input":{}, "output":{"Status":"int"}},
    "debloquer": {"input":{}, "output":{"Status":"int"}},
    "fermerSession": {"input":{}, "output":{"Status":"int"}},
    "ouvrirSession": {"input":{}, "output":{"Status":"int"}},
    "payer": {"input":{}, "output":{"Status":"int"}},
    "scanner": {"input":{}, "output":{"Status":"int"}},
    #"supprimer": {"input":{}, "output":{"Status":"int"}},
    "transmission": {"input":{}, "output":{"Status":"int"}}
    }
    
# %% Some largish clusters that are missing system tests.

missing = [3, 4, 5]

# %% Build models for those clusters and evaluate the models.
for n in missing:
    cluster = agilkia.TraceSet(cust.get_cluster(n))
    print(f"========== cluster {n} has {len(cluster)} traces ============")

    # Learn a test-generation model for this cluster.
    ex = agilkia.TracePrefixExtractor()
    X = ex.fit_transform(cluster)
    y = ex.get_labels()
    #print(X.head())
    #print(f"y: {y[0:20]}...")

    # Evaluate various classifiers
    for name,alg in [
                ("Tree", DecisionTreeClassifier()),   # fast, 0.951
                ("GBC", GradientBoostingClassifier()),  # slower, 0.951
                ("RandForest", RandomForestClassifier(n_estimators=100)),  # med 0.951
                ("AdaBoost", AdaBoostClassifier()),  # 0.421, some labels have no predictions
                #("Gaussian", GaussianProcessClassifier(max_iter_predict=10)),   # VERY slow, 0.886
                ("NeuralNet", MLPClassifier(solver='lbfgs')),  # adam solver doesn't converge. 0.924
                ("KNeighbors", KNeighborsClassifier()),  # fast, 0.948
                ("NaiveBayes", GaussianNB()),  # fast, F1 undef. 0.839
                ("LinearSVC", LinearSVC()),  # fast, 0.886
                ("Dummy", DummyClassifier()),  # fast, 0.130
                ("LogReg", LogisticRegression(solver='lbfgs', max_iter=200, multi_class='auto'))  # med 0.89
            ]:
        pipe = Pipeline([
            ("Normalize", MinMaxScaler()),
            (name, alg)
            ])
        scores = cross_val_score(pipe, X, y, cv=10, scoring='f1_macro')
        # print(scores)
        print(f"{name:20s} & {scores.mean():0.3f} (+/- {scores.std() * 2:0.3f})\\\\")


# %% Now use Decision Tree model to generate some tests.
        
def gen_tests_for(traceset, name, traces=5):
    print(f"========== {name} has {len(traceset)} traces ============")

    # Learn a test-generation model for this cluster.
    ex = agilkia.TracePrefixExtractor()
    X = ex.fit_transform(traceset)
    y = ex.get_labels()
    print(f"  it has {len(X)} trace prefixes")
    # Train a decision tree model on this cluster
    model = Pipeline([
        ("Extractor", ex),
        ("Normalize", MinMaxScaler()),
        ("Tree", DecisionTreeClassifier())   # fast, 0.951
        ])
    model.fit(traceset, y)
    
    rand = random.Random(1234)
    smart = agilkia.SmartSequenceGenerator([], method_signatures=signature, rand=rand)
    smart.trace_set.set_event_chars(traceset.get_event_chars())
    # generate some tests
    for i in range(traces):
        smart.generate_trace_with_model(model, length=100)
    for tr in smart.trace_set:
        print(f"    {tr}")
    return model

# %%

for n in missing:
    cluster = agilkia.TraceSet(cust.get_cluster(n))
    gen_tests_for(cluster, f"cluster {n}")

# %%
    
model = gen_tests_for(cust, "all customer traces", traces=30)

# %% Get frequency of a generated trace (maybe partial)

max_length = 35
chars = cust.get_event_chars()


# %%

def freq(trace:agilkia.Trace) -> float:
    return trace.meta_data["freq"]
    
def gen_all():
    smart = agilkia.SmartSequenceGenerator(urls=[], method_signatures=signature, 
                                           verbose=False, action_chars=chars)
    seqs = smart.generate_all_traces(model, length=max_length, action_prob=0.01, path_prob=0.01,
                                     partial=True)
    return seqs

# %% Time the test suite generation
    
duration = timeit.timeit(gen_all, number=10) / 10.0
print(f"duration = {duration} seconds")

# %%

seqs = gen_all()
sorted_seqs = sorted(seqs, key=freq, reverse=True)
print(f"generated {len(seqs)} sequences")
total = 0.0
for tr in sorted_seqs:
    etc = "" if len(tr) < max_length else "???"
    percent = freq(tr) * 100
    total += percent
    print(f"  {percent:5.2f}% {tr.to_string(to_char=chars)}{etc}")
print(f" {total:6.2f}% total")

# %%

ex = model[0]
tree = model[-1]
print(tree)

# %% 

# Export as dot file
export_graphviz(tree, out_file='tree.dot', 
                feature_names = ex.get_feature_names(),
                class_names = model.classes_,
                rounded = True, proportion = False, 
                precision = 2, filled = True)

# %% print as text
tree_str = export_text(tree, feature_names=ex.get_feature_names())
print(tree_str)

# %%

if not TEST_MODE:
    # Convert to png using system command (requires Graphviz)
    from subprocess import call
    call(['dot', '-Tpng', 'tree.dot', '-o', 'tree.png', '-Gdpi=300'])

    # Display in jupyter notebook (does not work in Spyder)
    from IPython.display import Image
    Image(filename = 'tree.png')



"""
Simple random test generator for SOAP web services.

Author: Mark Utting, 2019

TODO:
* provide a way of generating related inputs like (lat,long) together.
* improve SmartSequenceGenerator to pass scikit-learn estimator tests?
  See sklearn.utils.estimator_checks.check_estimator(Estimator)

"""


import csv
import requests
import zeep   # type: ignore
import zeep.helpers   # type: ignore
import getpass
import operator
import random
import collections
import unittest
import sklearn.base
import pandas as pd
from sklearn.utils.validation import check_is_fitted
from pathlib import Path
from pprint import pprint
from typing import Tuple, List, Mapping, Dict, Any, Optional, Union

from . json_traces import Event, Trace, TraceSet, TraceEncoder, MetaData


# A signature of a method maps "input"/"output" to the dictionary of input/output names and types.
Signature = Mapping[str, Mapping[str, str]]
InputRules = Dict[str, List[str]]


# TODO: make these user-configurable
DUMP_WSDL = False         # save each *.wsdl file into current directory.
DUMP_SIGNATURES = False    # save summary of methods into *_signatures.txt
GOOD_PASSWORD = "<GOOD_PASSWORD>"
TRACE_END = "<END>"


def read_input_rules(file: Path) -> InputRules:
    """Reads a CSV file of input values.

    The header line of the CSV file should contain headers: Name,Frequency,Value.
    (but the Frequency column is optional, and missing frequencies default to 1).

    For example if one line contains 'size,3,100' and another contains 'size,2,200',
    then the resulting input rules will define a 3/5 chance of size being 100,
    and a 2/5 chance of it being 200.
    """
    input_rules: InputRules = {}
    with open(file, "r") as input:
        for row in csv.DictReader(input):
            name = row["Name"]
            freq = row.get("Frequency", "")
            freq_int = int(freq) if freq else 1
            value = row["Value"]
            value_list = input_rules.get(name, [])
            for i in range(freq_int):
                value_list.append(value)
            input_rules[name] = value_list  # update it after appending new values
    print(input_rules)
    return input_rules


def uniq(d):
    """Returns the unique value of a dictionary, else an empty dictionary."""
    result = {}
    for k, v in d.items():
        if result == {} or result == v:
            result = v
            return result  # temp hack - ITM 3 ports have slight differences.
        else:
            print(f"WARNING: uniq sees different values.\n" +
                  " val1={result}\n  val2={v}")
            return {}
    return result


class TestUniq(unittest.TestCase):
    """Some unit tests of the uniq function."""

    def test_normal(self):
        self.assertEqual("def", uniq({"abc": "def"}))

    # TODO: assert uniq({"abc":"one", "xyz":"two"}) == {}

    def test_duplicate_values(self):
        self.assertEquals("one", uniq({"abc": "one", "xyz": "one"}))


def parse_elements(elements):
    """Helper function for build_interface."""
    all_elements = {}
    for name, element in elements:
        all_elements[name] = {}
        all_elements[name]['optional'] = element.is_optional
        if hasattr(element.type, 'elements'):
            all_elements[name]['type'] = parse_elements(element.type.elements)
        else:
            all_elements[name]['type'] = str(element.type)
    return all_elements


def build_interface(client: zeep.Client) -> Dict[str, Dict[str, Any]]:
    """Returns a nested dictionary structure for the methods of client.

    Typical usage to get a method called "Login" is:
    ```build_interface(client)[service][port]["operations"]["Login"]```
    """
    interface: Dict[str, Dict[str, Any]] = {}
    for service in client.wsdl.services.values():
        interface[service.name] = {}
        for port in service.ports.values():
            interface[service.name][port.name] = {}
            operations: Dict[str, Any] = {}
            for operation in port.binding._operations.values():
                operations[operation.name] = {}
                operations[operation.name]['input'] = {}
                elements = operation.input.body.type.elements
                operations[operation.name]['input'] = parse_elements(elements)
            interface[service.name][port.name]['operations'] = operations
    return interface


def print_signatures(client: zeep.Client, out):
    """Print a short summary of each operation signature offered by client."""
    # From: https://stackoverflow.com/questions/50089400/introspecting-a-wsdl-with-python-zeep
    for service in client.wsdl.services.values():
        out.write(f"service: {service.name}\n")
        for port in service.ports.values():
            out.write(f"  port: {port.name}\n")
            operations = sorted(
                port.binding._operations.values(),
                key=operator.attrgetter('name'))
            for operation in operations:
                action = operation.name
                inputs = operation.input.signature()
                outputs = operation.output.signature()
                out.write(f"    {action}({inputs})  --> ({outputs})\n")


class RandomTester:
    """Does random testing of a given web service.

    Give it a URL to a web service (or a list of URLs if there are several web services),
    and it will read the WSDL specifications from those web services and
    generate any number of random test sequences to test the methods.

    For more sophisticated (user-directed) testing you can also:
    * supply a username and password if login credentials are needed.
    * supply the subset of method names that you want to focus on testing (default is all).
    * supply a set of default input values (or generation functions) for each data type.
    * supply a set of input values (or generation functions) for each named input parameter.
    """
    def __init__(self,
                 urls: Union[str, List[str]],
                 methods_to_test: List[str] = None,
                 input_rules: Dict[str, List] = None,
                 rand: random.Random = None,
                 action_chars: Mapping[str, str] = None,
                 verbose: bool = False):
        """Creates a random tester for the server url and set of web services on that server.

        Args:
            urls (str or List[str]): URLs to the web services, used to find the WSDL files.
            methods_to_test (List[str]): only these methods will be tested (None means all).
            input_rules (Dict[str,List]): maps each input parameter name to a list of
                possible values, one of which will be chosen randomly.
            rand (random.Random): the random number generator used to generate tests.
            action_chars (Mapping[str, str]): optional action-to-character map, for visualisation.
            verbose (bool): True means print progress messages during test generation.
        """
        self.urls = [urls] if isinstance(urls, str) else urls
        self.username: Optional[str] = None
        self.password: Optional[str] = None
        self.random = random.Random() if rand is None else rand
        self.verbose = verbose
        self.clients_and_methods: List[Tuple[zeep.Service, Dict[str, Signature]]] = []
        self.methods_to_test = methods_to_test
        self.methods_allowed = [] if methods_to_test is None else methods_to_test
        # maps each parameter to list of possible 'values'
        self.named_input_rules = {} if input_rules is None else input_rules
        meta = TraceSet.get_default_meta_data()
        meta["source"] = "RandomTester"
        meta["web_services"] = self.urls
        meta["methods_to_test"] = methods_to_test
        meta["input_rules"] = input_rules
        meta["method_signatures"] = {}  # see add_web_service
        meta["action_chars"] = action_chars
        new_trace = Trace([], random_state=self.random.getstate())
        self.curr_events = new_trace.events  # mutable list to append to.
        self.trace_set = TraceSet([], meta)
        self.trace_set.append(new_trace)
        for w in self.urls:
            self.add_web_service(w)

    def set_username(self, username: str, password: str = None):
        """Set the username and (optional) password to be used for the subsequent operations.
        If password is not supplied, this method will immediately interactively prompt for it.
        """
        self.username = username
        self.trace_set.meta_data["username"] = username
        self.password = password or getpass.getpass(f"Please enter password for user {username}:")

    def add_web_service(self, url: str):
        """Add another web service using the given url."""
        wsdl = url + ("" if url.upper().endswith("WSDL") else ".asmx?WSDL")
        name = url.split("/")[-1]
        print("  loading WSDL: ", wsdl)
        if DUMP_WSDL:
            # save the WSDL for reference
            r = requests.get(wsdl, allow_redirects=True)
            open(f"{name}.wsdl", 'wb').write(r.content)
        # now create the client interface for this web service
        client = zeep.Client(wsdl=wsdl)
        interface = build_interface(client)
        pprint([(k, len(v["operations"])) for k, v in uniq(interface).items()])
        if DUMP_SIGNATURES:
            # save summary of this web service into a signatures file
            with open(f"{name}_signatures.txt", "w") as sig:
                print_signatures(client, sig)
        if not uniq(interface):
            print(f"WARNING: web service {name} has empty interface?")
            pprint(interface)
        else:
            ops = uniq(uniq(interface))["operations"]
            self.clients_and_methods.append((client, ops))
            self.trace_set.meta_data["method_signatures"].update(ops)
            if self.methods_to_test is None:
                self.methods_allowed += list(ops.keys())

    def _find_method(self, name: str) -> Tuple[zeep.Client, Signature]:
        """Find the given method in one of the web services and returns its signature."""
        for (client, interface) in self.clients_and_methods:
            if name in interface:
                return client, interface[name]
        raise Exception(f"could not find {name} in any WSDL specifications.")

    def choose_input_value(self, arg_name: str) -> str:
        """Choose an appropriate value for the input argument called 'arg_name'.
        If no set of input rules is defined for 'arg_name', then 'generate_input_value'
        is called to generate a suitable input value.  Subclasses can override this.

        Args:
            arg_name (str): the name of the input parameter.

        Returns:
            a string if successful, or None if no suitable value was found.
        """
        values = self.named_input_rules.get(arg_name, None)
        if values is None:
            return self.generate_input_value(arg_name)
        val = self.random.choice(values)
        return val

    def generate_input_value(self, arg_name: str) -> Any:
        """Can be overridden in subclasses to generate smart values for an input argument."""
        print(f"ERROR: please define possible parameter values for input {arg_name}")
        return None

    def _insert_password(self, arg_value: str) -> str:
        if arg_value == GOOD_PASSWORD:
            if self.password is None:
                raise Exception("Please call set_username before using " + GOOD_PASSWORD)
            return self.password
        else:
            return arg_value

    def get_methods(self) -> Mapping[str, Signature]:
        """Return the set of all method names in all the web services."""
        methods = {}
        for (client, interface) in self.clients_and_methods:
            methods.update(interface)
        return methods

    def summary(self, value) -> str:
        """Returns a one-line summary of the given value."""
        s = str(value).replace("\n", "").replace(" ", "")
        return s[:95]

    def decode_outputs(self, raw) -> Dict[str, Any]:
        """Decode the outputs from a web service/site call into a dictionary.
        
        This adds a 'Status' entry in the output dictionary,
        to say if the operation was successful (0) or not (non-zero).
        """
        # Since raw comes from a zeep call, it should be XML.
        if isinstance(raw, dict):
            out = raw.copy()
        elif isinstance(raw, str):
            out = {"value": raw}
        else:
            out = TraceEncoder().default(raw)
            if not isinstance(out, dict):
                out = {"value": out}
        if "Status" not in out:
            out["Status"] = 0
        return out

    def call_method(self, name: str, args: Dict[str, Any] = None,
                    meta_data: Optional[MetaData] = None):
        """Call the web service name(args) and add the result to trace.

        Args:
            name (str): the name of the method to call.
            args (dict): the input values for the method.  If args=None, then this method uses
                'choose_input_value' to choose appropriate values for each argument
                value of the method.
            meta_data: optional meta data to add to the resulting Event.
        Returns:
        Before the call, this method replaces some symbolic inputs by actual concrete values.
        For example the correct password token is replaced by the real password --
        this avoids recording the real password in the inputs of the trace.

        Returns:
            The whole Event object created by this method call.
        """
        (client, signature) = self._find_method(name)
        inputs = signature["input"]
        if args is None or (len(args) == 0 and len(inputs) > 0):
            args = {n: self.choose_input_value(n) for n in inputs.keys()}
        if None in args.values():
            print(f"skipping method {name}.  Please define missing input values.")
            return None
        if self.verbose:
            print(f"    call {name}{args}")
        if client is None:
            out = {"Status": 0}  # dummy results, always succeeds.
        else:
            # insert special secret argument values if requested
            args_list = [self._insert_password(arg) for (n, arg) in args.items()]
            raw_out = getattr(client.service, name)(*args_list)
            out = self.decode_outputs(raw_out)
        event = Event(name, args, out, meta_data=meta_data)
        self.curr_events.append(event)
        if self.verbose:
            print(f"      -> {self.summary(event.outputs)}")
        return event

    def generate_trace(self, start=True, length=20, methods: List[str] = None) -> Trace:
        """Generates the requested length of test steps, choosing methods at random.

        Args:
            start (bool): True means that a new trace is started (unless current one is empty).
            length (int): The number of steps to generate (default=20).
            methods (List[str]): only these methods will be chosen (None means all are allowed)

        Returns:
            the whole of the current trace that has been generated so far.
        """
        if start:
            if len(self.curr_events) > 0:
                new_trace = Trace([], random_state=self.random.getstate())
                self.curr_events = new_trace.events  # mutable list to append to.
                self.trace_set.append(new_trace)
        if methods is None:
            methods = self.methods_allowed
        for i in range(length):  # TODO: continue while Status==0?
            self.call_method(self.random.choice(methods))
        return self.trace_set.traces[-1]


class TracePrefixExtractor(sklearn.base.BaseEstimator, sklearn.base.TransformerMixin):
    """Extracts prefixes of traces so that the next action can be predicted.
    
    Attributes:
        vocabulary_ (dict): A dictionary mapping feature names to feature indices.
        feature_names_ (list): A list of length n_features containing the feature names.
    """

    def __init__(self):
        self._traces = None
        pass

    def get_feature_names(self):
        """Gets the list of column names for the generated data tables."""
        check_is_fitted(self, 'is_fitted_')
        return self.feature_names_

    def set_feature_names(self, names:List[str]):
        """Sets self.feature_names_ to the given list of feature names.
        And sets self.vocabulary_ to the inverse mapping - names to position.
        """
        self.feature_names_ = names
        self.vocabulary_ = {(col,i) for (i,col) in enumerate(self.feature_names_)}

    def get_prefix_features(self, events:List[Event]) -> Dict[str, float]:
        """Converts a sequence of events into numeric training data to predict next action.
        
        Subclasses could override this to change the feature encoding.
        (If they want to change the feature names, they should also override set_feature_names)
        """
        counts = collections.Counter([ev.action for ev in events])
        # TODO: add 'recent' event counts if that is desired.
        return counts

    def fit(self, X:TraceSet, y=None):
        """Fit remembers the given TraceSet and calculates the feature names.

        Args:
            X (TraceSet): the set of traces to be fitted.
            y (None): unused.

        Returns:
            self

        Note that fit() must be called before transform() or get_feature_names().
        """
        if not isinstance(X, TraceSet):
            raise Exception("TracePrefixExtractor.fit input must be TraceSet.")
        self._traceset = X
        self.set_feature_names(X.get_all_actions())
        self.is_fitted_ = True
        return self

    def transform(self, X:Union[TraceSet, List[Event]]):
        """Transforms a set of traces, or an event sequence, into a Pandas DataFrame
        of training data.  Note that the columns of the resulting DataFrame are fixed
        during the fit() method, so any new kinds of actions appearing in this X
        input will be ignored.

        There are two different behaviors, depending upon the input type of X.
            * if traces is a TraceSet, all prefixes of all traces are converted into
              training data, and the corresponding expected y labels (action name) for
              all those prefixes are available from get_labels().
            * if traces is a list of events, then the result will contain just
              a single row, which will be the data for that whole trace.
        """
        check_is_fitted(self, 'is_fitted_')
        data = []
        self._y = []
        if isinstance(X, TraceSet):
            for tr in X.traces:
                for size in range(len(tr) + 1):  # every prefix, plus the whole trace.
                    events = tr.events[0:size]
                    features = self.get_prefix_features(events)
                    action = tr[size].action if size < len(tr) else TRACE_END
                    data.append(features)
                    self._y.append(action)
        elif isinstance(X, list) and (X == [] or isinstance(X[0], Event)):
            features = self.get_prefix_features(X)
            data = [features]
            self._y = [TRACE_END]
        else:
            raise Exception("TracePrefixExtractor.transform input must be TraceSet or [Events].")
        df = pd.DataFrame(data, columns=self.feature_names_)
        df.fillna(0, inplace=True)
        return df

    def get_labels(self):
        """Get the output labels (action names) corresponding to the last transform() call."""
        check_is_fitted(self, 'is_fitted_')
        return self._y


class SmartSequenceGenerator(RandomTester):
    """Generates test sequences from an ML model that suggests what actions can come next."""
    
    def __init__(self,
                 urls: Union[str, List[str]],
                 method_signatures:Dict[str, Signature] = None,
                 methods_to_test: List[str] = None,
                 input_rules: Dict[str, List] = None,
                 rand: random.Random = None,
                 action_chars: Mapping[str, str] = None,
                 verbose: bool = False):
        """A test sequence generator that uses a machine learning model to predict next action.
        
        Args:
            urls: Optional list of WSDL URLs.  If this is the empty list, then
                method_signatures must be provided.
            method_signatures: Optional mapping from method names to input/output signatures.
                This will be inferred automatically if urls is provided.
            methods_to_test (List[str]): only these methods will be tested (None means all).
            input_rules (Dict[str,List]): maps each input parameter name to a list of
                possible values, one of which will be chosen randomly.
            rand (random.Random): the random number generator used to generate tests.
            action_chars (Mapping[str, str]): optional action-to-character map, for visualisation.
            verbose (bool): True means print progress messages during test generation.
        """
        if methods_to_test is None and method_signatures is not None:
            methods_to_test = list(method_signatures.keys())
        super().__init__(urls, methods_to_test=methods_to_test, input_rules=input_rules,
                         rand=rand, action_chars=action_chars, verbose=verbose)
        # Quick hack to get some dummy signatures set up when running offline (no web service)
        if not urls:
            self.clients_and_methods.append((None, method_signatures))
            self.trace_set.meta_data["method_signatures"].update(method_signatures)
        if self.methods_to_test is None:
            self.methods_allowed += sorted(list(method_signatures.keys()))

    def generate_trace_with_model(self, model, start=True, length=20):
        """Generates one sequence test steps, choosing actions using the given model.
        The generated trace terminates either when the model says <end> or after length steps.

        Args:
            model (Classifier): ML model that takes an Event list and predicts next action name.
            start (bool): True means that a new trace is started, beginning with a "Login" call.
            length (int): The maximum number of steps to generate in one trace (default=20).

        Returns:
            the whole of the current trace that has been generated so far.
        """
        # start a new (empty) trace if requested.
        self.generate_trace(start=start, length=0)
        result = self.curr_events
        for i in range(length):
            [proba] = model.predict_proba(self.curr_events)
            [action_num] = self.random.choices(range(len(proba)), proba) 
            action = model.classes_[action_num] # WAS: self.methods_allowed[action_num]
            if self.verbose:
                probs = ",".join([f"{int(p*100)}" for p in proba])
                print(f"{i:3d}: {action:20s} {probs}")
            if action == TRACE_END:
                self.generate_trace(start=True, length=0)
                break  # we view length as a maximum...
            else:
                self.call_method(action)
        return result

    def generate_all_traces(self, model, length=5, action_prob=0.01, path_prob=1.0e-12,
                            partial=True) -> List[Trace]:
        """Generate all traces that satisfy the given constraints.

        Args:
            model: the trained ML model used to predict the next action.
            length (int): maximum length of each generated trace.
            action_prob (float): only do actions with at least this probability.
            path_prob (float): only include paths with at least this total probability.
            partial (bool): True means include partial traces.  False gives complete traces only.

        Returns:
            A list of all the Trace objects that satisfy the given constraints.
                Note that all complete traces will have len(tr)<length, whereas all partial
                traces will have len(tr)==length.
        """
        results = []
        def depth_first_search(prefix, prob):
            indent = ">" * len(prefix)
            # print(indent + ",".join([ev.action for ev in prefix]))
            if len(prefix) >= length:
                if partial:
                    results.append(Trace(events=prefix, meta_data={"freq":prob}))
            else:
                [proba] = model.predict_proba(prefix)
                for i, p in enumerate(proba):
                    if p >= action_prob and prob * p >= path_prob:
                        action = model.classes_[i]
                        if action == TRACE_END:
                            results.append(Trace(events=prefix, meta_data={"freq":prob * p}))
                        else:
                            if self.verbose:
                                print(indent + f" trying {action}")
                            depth_first_search(prefix + [Event(action,{},{})], prob * p)
        depth_first_search([], 1.0)
        return results

    def execute_test(self, trace:Trace, max_retry:int=0):
        """Executes the given test trace and adds the resulting trace to this set.
        
        Note that if the given trace contains events with missing input values, then
        suitable input values will be generated using 'choose_input_value'.
        Progress messages will be printed if self.verbose is True.
        
        Args:
            trace: the trace to execute (with or without input values).
            max_retry: retry failed operations up to this number of times,
                choosing different random input values each time.
        """
        self.generate_trace(start=True, length=0)
        for ev in trace:
            for i in range(1 + max_retry):
                ev2 = self.call_method(ev.action, ev.inputs, meta_data=ev.meta_data)
                if ev2.status == 0:
                    break
                else:
                    if self.verbose and i < max_retry:
                        print(f"    retry {i+1} ...")


if __name__ == "__main__":
    unittest.main()




# -*- coding: utf-8 -*-
"""
Data structures for Traces and Sets of Traces.

This defines the 'Event', 'Trace' and 'TraceSet' classes, plus helper functions.

NOTES
=====
1.  *private data fields* (starting with '_') will not be stored in the JSON files.
    For example, each Trace object has a '_parent' point to its TraceSet, but this
    is not stored in the JSON file, since the hierarchical structure of the JSON
    already captures the parent-child relationship between TraceSet and Trace.
2.  *JSON-traces file version numbers* follow the usual Semantic Versioning scheme:
    *(Major.Minor.Patch)*.
    ``TraceSet.upgrade_json_data`` currently just prints a warning message when a
    program running older code reads a JSON file with a newer MINOR version number.
    This allows graceful updating of one program at a time, but does
    have the danger that a older program may read newer data (with a warning),
    then SAVE that data in the slightly older format, thus losing some data.
    But a strict version-equality means that all programs have to be updated
    simultaneously, which is a pain.

Ideas / Tasks to do
===================
    * DONE: add save_as_arff() method like to_pandas.
    * DONE: store event_chars into meta_data.
    * DONE: store signatures into meta_data.
    * DONE: create Event class and make it dict-like.
    * DONE: add support for splitting traces into 'sessions' via splitting or grouping.
    * DONE: add support for clustering traces
    * DONE: add support for visualising the clusters (TSNE).
    * DONE: add 'meta_data' to Trace and Event objects too (replace properties)
    * DONE: add unit tests for clustering...  (Note: not saved in JSON!)
    * DONE: split RandomTester into SmartSequenceGenerator subclass (better meta-data).
    * provide an easy way of copying meta-data across, or cloning a Trace[Set] with new traces.
    * extend to_pandas() to allow user-defined columns to be added.
    * split the test execution methods out of RandomTester into a delegate class,
        so that we can generate tests without executing them, and execute without generating.

@author: utting@usc.edu.au
"""

import os
import sys
from pathlib import Path  # object-oriented filenames!
from collections import defaultdict
import json
import decimal
import datetime
import re
import xml.etree.ElementTree as ET
import pandas as pd            # type: ignore
import numpy as np             # type: ignore
import sklearn.cluster         # type: ignore
import sklearn.preprocessing   # type: ignore
import matplotlib.pyplot as plt
import matplotlib.cm as pltcm
from sklearn.manifold import TSNE
# liac-arff from https://pypi.org/project/liac-arff (via pip)
# import arff                    # type: ignore
from typing import List, Set, Mapping, Dict, Union, Any, Optional, cast


TRACE_SET_VERSION = "0.1.4"

MetaData = Dict[str, Any]


def safe_name(string: str) -> str:
    """Returns 'string' with all non-alpha-numeric characters replaced by '_'."""
    return re.sub("[^A-Za-z0-9]", "_", string)


class Event:
    """An Event is a dictionary-like object that records all the details of an event.

    Public data fields include:
        * self.action (str): the full action name.
        * self.inputs (Dict[str,Any]): the named inputs and their values.
        * self.outputs (Dict[str,Any]): the named outputs and their values.
        * self.meta_data (Dict[str,Any]): any extra properties such as "timestamp".
          Note: if self.meta_data["timestamp"] is present, it should be in ISO 8601 format.
          Or use get_meta(key) to get an individual meta-data value.
    """

    def __init__(self, action: str, inputs: Dict[str, Any], outputs: Dict[str, Any],
                 meta_data: Optional[MetaData] = None):
        self.action = action
        self.inputs = inputs
        self.outputs = outputs
        self.meta_data: MetaData = {} if meta_data is None else meta_data.copy()

    @property
    def status(self) -> int:
        """Read-only status of the operation, where 0 means success.
        If no output 'Status' is available, this method always returns 0.
        """
        return int(self.outputs.get("Status", "0"))

    @property
    def error_message(self) -> str:
        """Read-only error message output by this operation.
        If no output['Error'] field is available, this method always returns "".
        """
        return self.outputs.get("Error", "")

    def __str__(self):
        """Shows action, inputs and outputs, but elides any meta-data."""
        return f"Event({self.action}, {self.inputs}, {self.outputs})"


class Trace:
    """Represents a single trace, which contains a sequence of events.

    Public data fields include:
        * self.events: List[Event].  However, the iteration, indexing, and len(_) methods
          have been lifted from the events list up to this Trace object, so you
          may not need to access self.events at all.
        * self.meta_data: MetaData.  Or use get_meta(key) to get an individual meta-data value.
    """

    def __init__(self, events: List[Event], parent: 'TraceSet' = None,
                 meta_data: Optional[MetaData] = None,
                 random_state=None):
        """Create a Trace object from a list of events.

        Args:
            events: the sequence of Events that make up this trace.
            parent: the TraceSet that this trace is part of.
            meta_data: any meta-data associated with this individual trace.
            random_state: If this trace was generated using some randomness, you should supply
                this optional parameter, to record the state of the random generator at the
                start of the sequence.  For example, rand_state=rand.getstate().
                If this value is supplied, it is added into the meta-data.
        """
        if events and not isinstance(events[0], Event):
            raise Exception("Events required, not: " + str(events[0]) + " ...")
        self.events = events
        self._parent = parent
        self.meta_data: MetaData = {} if meta_data is None else meta_data
        if random_state is not None:
            self.meta_data["random_state"] = random_state

    def trace_set(self):
        """Returns the TraceSet that this trace is part of, or None if not known."""
        return self._parent

    def __iter__(self):
        return self.events.__iter__()

    def __len__(self):
        return len(self.events)

    def __getitem__(self, key):
        return self.events[key]

    def get_meta(self, key: str) -> Optional[Any]:
        """Returns requested meta data, or None if that key does not exist."""
        if key in self.meta_data:
            return self.meta_data[key]
        else:
            return None

    def append(self, event: Event):
        if not isinstance(event, Event):
            raise Exception("Event required, not: " + str(event))
        self.events.append(event)

    def action_counts(self) -> Dict[str, int]:
        """Returns a dictionary of how many times each action occurs in this trace.

        Returns:
            A dictionary of counts that can be used for clustering traces.
        """
        result = defaultdict(int)
        for ev in self.events:
            result[ev.action] += 1
        return result

    def action_status_counts(self) -> Dict[str, int]:
        """Counts how many times each action-status pair occurs in this trace.

        Returns:
            A dictionary of counts that can be used for clustering traces.
        """
        result = defaultdict(int)
        for ev in self.events:
            key = ev.action + "_" + str(ev.status)
            result[key] += 1
        return result

    def to_string(self,
                  to_char: Dict[str, str] = None,
                  compress: List[str] = None,
                  color_status: bool = False):
        """Return a one-line summary of this trace, one character per event.
        See 'trace_to_string' for details.
        NOTE: throws an exception if no to_char map is given and this trace has no parent.
        """
        if to_char is None:
            if self._parent is None:
                raise Exception("Cannot view trace with no parent and no to_char map.")
            to_char = self._parent.get_event_chars()
        return trace_to_string(self.events, to_char, compress=compress, color_status=color_status)

    def __str__(self):
        try:
            return self.to_string()
        except Exception:
            return "???"


class TraceSet:
    """Represents a set of traces, either generated or recorded.

    Invariants:
        * forall tr:self.traces (tr._parent is self)
          (TODO: set _parent to None when a trace is removed?)
        * self.meta_data is a dict with keys: date, source at least.

    Public data fields include:
        * self.traces: List[Trace].  However, the iteration, indexing, and len(_) methods
          have been lifted from the trace list up to the top-level TraceSet object, so you
          may not need to access self.traces at all.
        * self.meta_data: MetaData.  Or use get_meta(key) to get an individual meta-data value.
        * self.version: str.  Version number of this TraceSet object.
    """

    meta_data: MetaData
    _event_chars: Optional[Dict[str, str]]

    def __init__(self, traces: List[Trace], meta_data: Dict[str, Any] = None):
        """Create a TraceSet object from a list of Traces.

        Args:
            traces: the list of Traces.
            meta_data: a dictionary that captures the important meta-data for this set of traces.
                If this TraceSet is to be saved into a file, the meta-data should include at
                least the GDF (General Data Format) compulsory fields,
                which are based on the Dublin Core:
                    "date" in ISO 8601 format;
                    "dataset" for the official name of this TraceSet;
                    "source" for the origin of the dataset;
                    and any other meta-data that is available.
        """
        self.version = TRACE_SET_VERSION
        self.traces = traces
        self._clusters: List[int] = None
        self._cluster_data: pd.DataFrame = None
        trace_parents = set()
        # add all the trace to this set.
        for tr in self.traces:
            if isinstance(tr, Trace):
                if tr._parent and tr._parent != self:
                    trace_parents.add(tr._parent)
                tr._parent = self
            else:
                raise Exception("TraceSet expects List[Trace], not: " + str(type(tr)))
        if meta_data is None:
            if len(trace_parents) == 1:
                # copy across meta-data, since all traces come from same place.
                self.meta_data = next(iter(trace_parents)).meta_data.copy()
                # print("DEBUG: copyied meta-data from traces:", self.meta_data)
                now = datetime.datetime.now().isoformat()
                self.meta_data["date"] = now
            else:
                self.meta_data = self.get_default_meta_data()
        else:
            self.meta_data = meta_data.copy()
        self._event_chars = None  # recalculated if set of traces grows.

    def __iter__(self):
        return self.traces.__iter__()

    def __len__(self):
        return len(self.traces)

    def __getitem__(self, key):
        return self.traces[key]

    def message(self, msg: str):
        """Print a progress message."""
        print("   ", msg)

    @classmethod
    def get_default_meta_data(cls) -> Dict[str, Any]:
        """Generates some basic meta-data such as date, user and command line."""
        now = datetime.datetime.now().isoformat()
        user = os.path.expanduser('~').split('/')[-1]  # usually correct, but can be tricked.
        meta_data: Dict[str, Any] = {
                "date": now,
                "author": user,
                "dataset": "unknown",
                "action_chars": None
                }
        if len(sys.argv) > 0:
            meta_data["source"] = sys.argv[0]  # the path to the running script/tool.
            meta_data["cmdline"] = sys.argv
        return meta_data

    def get_meta(self, key: str) -> Optional[Any]:
        """Returns requested meta data, or None if that key does not exist."""
        if key in self.meta_data:
            return self.meta_data[key]
        else:
            return None

    def set_meta(self, key: str, value:Any) -> Optional[Any]:
        """Sets the requested meta data, and returns the old value if any."""
        old = None
        if key in self.meta_data:
            old = self.meta_data[key]
        self.meta_data[key] = value
        return old

    def append(self, trace: Trace):
        """Appends the given trace into this set of traces.
        This also sets its parent to be this trace set.
        """
        if not isinstance(trace, Trace):
            raise Exception("Trace required, not: " + str(trace))
        trace._parent = self
        self.traces.append(trace)
        self._event_chars = None  # we will recalculate this later

    def set_event_chars(self, given: Mapping[str, str] = None):
        """Sets up the event-to-char map that is used to visualise traces.

        This will calculate a default mapping for any actions that are not in given.
        (See 'default_map_to_chars').

        Args:
            given: optional pre-allocation of a few action names to chars.
            For good readability of the printed traces, it is recommended that extremely
            common actions should be mapped to 'small' characters like '.' or ','.
        """
        if given is None:
            new_given = self.get_meta("action_chars")
        else:
            self.meta_data["action_chars"] = given  # override any previous given map.
            new_given = cast(Dict[str, str], given).copy()  # copy so we don't change orginal.
        actions = all_action_names(self.traces)
        self._event_chars = default_map_to_chars(actions, given=new_given)

    def get_event_chars(self):
        """Gets the event-to-char map that is used to visualise traces.
        This maps each action name to a single character.
        If set_event_chars has not been called, this getter will calculate and cache
        a default mapping from action names to characters.
        """
        if self._event_chars is None:
            self.set_event_chars()
        return self._event_chars

    def __str__(self):
        name = self.meta_data["dataset"]  # required meta data
        return f"TraceSet '{name}' with {len(self)} traces."

    def save_to_json(self, file: Path) -> None:
        if isinstance(file, str):
            print(f"WARNING: converting {file} to Path.  Please learn to speak pathlib.")
            file = Path(file)
        with file.open("w") as output:
            json.dump(self, output, indent=2, cls=TraceEncoder)

    @classmethod
    def load_from_json(cls, file: Path) -> 'TraceSet':
        if isinstance(file, str):
            print(f"WARNING: converting {file} to Path.  Please learn to speak pathlib.")
            file = Path(file)
        if not isinstance(file, Path):
            raise Exception(f"load_from_json requires Path, not {file} (type={type(file)})")
        # with open(filename, "r") as input:
        data = json.loads(file.read_text())
        # Now check version and upgrade if necessary.
        if isinstance(data, list):
            # this file was pre-TraceSet, so just a list of lists of events.
            mtime = datetime.datetime.fromtimestamp(file.stat().st_mtime).isoformat()
            meta = {"date": mtime, "dataset": file.name, "source": "Upgraded from version 0.1"}
            traces = cls([], meta)
            for ev_list in data:
                events = [cls._create_event_object("0.1", ev) for ev in ev_list]
                traces.append(Trace(events))
            return traces
        elif isinstance(data, dict) and data.get("__class__", None) == "TraceSet":
            return cls.upgrade_json_data(data)
        else:
            raise Exception("unknown JSON file format: " + str(data)[0:60])

    @classmethod
    def upgrade_json_data(cls, json_data: Dict) -> 'TraceSet':
        version = json_data["version"]
        if version.startswith("0.1."):
            # This JSON file is compatible with our code.
            # First, convert json_data dicts to Trace and TraceSet objects.
            traceset = TraceSet([], json_data["meta_data"])
            for tr_data in json_data["traces"]:
                traceset.append(cls._create_trace_object(version, tr_data))
            # Next, see if any more little updates are needed.
            if version in ["0.1.2", "0.1.3", TRACE_SET_VERSION]:
                pass  # nothing more to do.
            elif version == "0.1.1":
                # Move given_event_chars into meta_data["action_chars"]
                # Note: traceset["version"] has already been updated to the latest.
                traceset.meta_data["actions_chars"] = json_data["given_event_chars"]
            else:
                # The JSON must be from a newer 0.1.x version, so give a warning.
                print(f"WARNING: reading {version} TraceSet using {TRACE_SET_VERSION} code.")
                print(f"         Some data may be lost.  Please upgrade this program.")
            return traceset
        raise Exception(f"upgrade of TraceSet v{version} to v{TRACE_SET_VERSION} not supported.")

    @classmethod
    def _create_trace_object(cls, version: str, tr_data: Dict[str, Any]) -> Trace:
        assert tr_data["__class__"] == "Trace"
        meta = tr_data.get("meta_data", {})
        rand = tr_data.get("random_state", None)
        if rand is not None:
            meta["random_state"] = rand
        events = [cls._create_event_object(version, ev) for ev in tr_data["events"]]
        return Trace(events, meta_data=meta)

    @classmethod
    def _create_event_object(cls, version: str, ev: Dict[str, Any]) -> Event:
        special = ["action", "inputs", "outputs"]
        action = ev["action"]
        inputs = ev["inputs"]
        outputs = ev["outputs"]
        if version <= "0.1.2":
            props = {key: ev[key] for key in ev if key not in special}
        elif version <= "0.1.3":
            props = ev["properties"]
        else:
            props = ev["meta_data"]
        return Event(action, inputs, outputs, props)

    def to_pandas(self) -> pd.DataFrame:
        """Converts all the traces into a single Pandas DataFrame (one event/row).

        The first three columns are 'Trace' and 'Event' which give the number of the
        trace and the position of the event within that trace, and 'Action' which is
        the name of the action of the event.
        Each named input value is recorded in a separate column.
        For outputs, by default there are just 'Status' (int) and 'Error' (str) columns.
        """
        return traces_to_pandas(self.traces)

    def arff_type(self, pandas_type: str) -> Union[str, List[str]]:
        """Maps each Pandas data type to the closest ARFF type."""
        if pd.api.types.is_integer_dtype(pandas_type):
            return "INTEGER"
        if pd.api.types.is_float_dtype(pandas_type):
            return "REAL"
        if pd.api.types.is_bool_dtype(pandas_type):
            return ["False", "True"]
        return "STRING"
        # TODO: check column to see if NOMINAL is better?
        # raise Exception(f"do not know how to translate Pandas type {pandas_type} to ARFF.")

    def save_to_arff(self, file: Path, name=None) -> None:
        """Save all the events in all traces into an ARFF file for machine learning.

        Args:
            filename: the name of the file to save into.  Should end with '.arff'.
            name: optional relation name to identify this data inside the ARFF file.
                The default is the base name of 'file'.
        """
        if isinstance(file, str):
            print(f"WARNING: converting {file} to Path.  Please learn to speak pathlib.")
            file = Path(file)
        if name is None:
            name = file.stem
        data = self.to_pandas()
        attributes = [(n, self.arff_type(t)) for (n, t) in zip(data.columns, data.dtypes)]
        try:
            import arff
        except ImportError:
            print("Please install ARFF support before using save_to_arff.")
            print("It is a pip only package:  pip install liac-arff")
            return
        with file.open("w") as output:
            contents = {
                "relation": safe_name(name),
                "attributes": attributes,
                "data": data.values,  # [[tr] for tr in trace_summaries],
                "description": "Events from " + name
                }
            arff.dump(contents, output)

    def with_traces_split(self, start_action: str = None, input_name: str = None,
                          comparator=None) -> 'TraceSet':
        """Returns a new TraceSet with each trace in this set split into shorter traces.

        It accepts several split criteria, and will start a new trace whenever any
        of those criteria are true.  At least one criteria must be supplied.

        Args:
            start_action: the name of an action that starts a new trace.
            input_name: the name of an input.  Whenever the value of this input
                changes, then a new trace should be started.  Note that events
                with this input missing are ignored for this splitting criteria.

        Returns:
            a new TraceSet, usually with more traces and shorter traces.
        """
        if start_action is None and input_name is None:
            raise Exception("split_traces requires at least one split criteria.")
        traces2 = TraceSet([], self.meta_data)
        # TODO: update meta data with split info?
        for old in self.traces:
            curr_trace = Trace([])
            traces2.append(curr_trace)
            prev_input = None
            for event in old:
                input_value = event.inputs.get(input_name, None)
                input_changed = input_value != prev_input and input_value is not None
                if (event.action == start_action or input_changed) and len(curr_trace) > 0:
                    curr_trace = Trace([])
                    traces2.append(curr_trace)
                curr_trace.append(event)
                if input_value is not None:
                    prev_input = input_value
                # NOTE: we could check end_action here.
        return traces2

    def with_traces_grouped_by(self, name: str, property: bool = False) -> 'TraceSet':
        """Returns a new TraceSet with each trace grouped into shorter traces.

        It generates a new trace for each distinct value of the given input or property name.

        Args:
            name: the name of an input.  A new trace is started for each value of this input
                (or property).  Note that events with this value missing are totally discarded.
            property: True means group by the property called name, rather than an input.

        Returns:
            a new TraceSet, usually with more traces and shorter traces.
        """
        # TODO: update meta data with split info?
        traces2 = TraceSet([], self.meta_data)
        for old in self.traces:
            groups = defaultdict(list)  # for each value this stores a list of Events.
            for event in old:
                if property:
                    value = event.meta_data.get(name, None)
                else:
                    value = event.inputs.get(name, None)
                if value is not None:
                    groups[value].append(event)
            for event_list in groups.values():
                traces2.append(Trace(event_list))
        return traces2

    def get_all_actions(self):
        """Returns a sorted list (with duplicates removed) of all the keys in data."""
        actions = set()
        for tr in self.traces:
            for ev in tr.events:
                actions.add(ev.action)
        return sorted(list(actions))

    def get_trace_data(self, method: str = "action_counts",
                       columns:List[str]=None) -> pd.DataFrame:
        """Returns a Pandas table of statistics/data about each trace.

        This can gather data using any of the zero-parameter data-gathering methods
        of the Trace class.  Any missing data values are replaced by zeroes.

        Note: you can add more data-gathering methods by defining a subclass of Trace
        and using that subclass when you create Trace objects.

        Args:
            method: the name of one of the methods in the Trace objects.
                This method must return a Dict[str, number] for some kind of number.
                The default is the ``action_counts()`` method, which corresponds to the
                *bag-of-words* algorithm.
            columns: optional list of column names.  This can be used to reorder or remove
                or add columns.  (Any added columns will be filled with zeroes).

        Returns:
            A table of data that can be used for clustering or machine learning.
            If columns is not specified, the columns of the table will be in alphabetical order.
            The i'th row of the table is the data for the i'th trace in this set.
        """
        trace_data = [getattr(tr, method)() for tr in self.traces]
        if columns is None:
            # columns is the sorted list of the union of all keys in trace_data.
            keys = set()
            for d in trace_data:
                keys.update(set(d.keys()))
            columns = sorted(list(keys))
        data = pd.DataFrame(trace_data, columns=columns)
        data.fillna(value=0, inplace=True)
        return data

    def create_clusters(self, data: pd.DataFrame, algorithm=None,
                        normalizer=None, fit: bool = True) -> int:
        """Runs a clustering algorithm on the given data and remembers the clusters.

        Note that clustering results are viewed as transient, so are currently not saved
        into JSON files.  Clustering must be repeated after loading a TraceSet.

        Args:
            data: a Pandas DataFrame, typically from get_trace_data(), with the i'th row
                of the DataFrame being for the i'th trace in this set of traces.
            algorithm: a clustering algorithm (default is MeanShift()).
            normalizer: a normalization algorithm (default is MinMaxScaler).
            fit: True means fit the data into clusters, False means just predict clusters
                assuming that the algorithm and normalizer have already been trained.

        Returns:
            The number of clusters generated.
        """
        if algorithm is None:
            if not fit:
                raise Exception("You must supply pre-fitted algorithm when fit=False")
            algorithm = sklearn.cluster.MeanShift()
        if normalizer is None:
            if not fit:
                raise Exception("You must supply pre-fitted normalizer when fit=False")
            normalizer = sklearn.preprocessing.MinMaxScaler()
            # normalizer = sklearn.preprocessing.RobustScaler()

        alg_name = str(algorithm).split("(")[0]
        self.message(f"running {alg_name} on {len(data)} traces.")
        if fit:
            normalizer.fit(data)
        self._cluster_data = pd.DataFrame(normalizer.transform(data), columns=data.columns)
        if fit:
            algorithm.fit(self._cluster_data)
            self._clusters = algorithm.labels_
        else:
            print(" pre predict len=", len(algorithm.labels_))
            self._clusters = algorithm.predict(self._cluster_data)
            print("post predict len=", len(algorithm.labels_), len(self._clusters))
        return max(self._clusters) + 1

    def is_clustered(self) -> bool:
        return self._clusters is not None

    def get_clusters(self) -> List[int]:
        """Get the list of cluster numbers for each trace.

        Precondition: self.is_clustered()
        """
        return self._clusters

    def visualize_clusters(self, algorithm=None, fit: bool = True,
                           xlim=None, ylim=None, cmap=None,
                           markers=None, markersize=None,
                           filename:str = None, block:bool=True):
        """Visualize the clusters from create_clusters().

        Args:
            algorithm: the visualization algorithm to map data into 2D (default TSNE).
            fit: True means fit the data, False means algorithm is pre-trained, so use it
                to just transform the data into 2D without fitting the data first.
                Note that TSNE does not support fit=False yet.
                If you want fit=False, use another dimension-reduction algorithm like PCA(...).
            xlim (Pair[float,float]): optional axis limits for the X axis.
            ylim (Pair[float,float]): optional axis limits for the Y axis.
            cmap (Union[ColorMap,str]): optional color map for the cluster colors,
                or the name of a color map.
                See https://matplotlib.org/3.1.1/tutorials/colors/colormaps.html.
                Default is 'brg', which has a wide range of
                colors going from blue through red to green, and prints in black and white
                okay - though very non-linear - because it does not go all the way to white.
            markers (matplotlib.markers.MarkerStyle): optional marker styles for clusters.
                If this is a string, then the i'th character in the string will be used for
                the i'th marker style.  See https://matplotlib.org/3.1.1/api/markers_api.html
                for the available marker characters.  Note that clusters will be drawn from 0
                up to n-1, so later clusters will be on top.  Also, the earlier clusters tend
                to have more elements.  One approach to improve readability is to use line-based
                shapes (from "1234+x|_") for the first few clusters (which have many points),
                and then filled shapes (from ".o<^>vsphPXd*") for the later clusters
                (which have few points).  Note also that you can use a space for the marker
                character of a cluster if you want to not display that cluster at all.
                However, if your markers string is shorter than the number of clusters,
                all remaining clusters will be displayed using the "o" marker.
            markersize (float): size of the markers in points (only when markers is a str).
                The default seems to be about 6 points.
            filename (str): optional file name to save image into, as well as displaying it.
            block (bool): True (the default) means wait for user to close figure before
                returning.  False means non-blocking.

            Limitations: if you call this multiple times with different numbers of clusters,
                the color map will not be exactly the same.
        """
        data = self._cluster_data
        if data is None or self._clusters is None:
            raise Exception("You must call create_clusters() before visualizing them!")
        num_clusters = max(self._clusters) + 1
        if algorithm is None:
            if not fit:
                raise Exception("You must supply pre-fitted algorithm when fit=False")
            algorithm = TSNE()
        alg_name = str(algorithm).split("(")[0]
        self.message(f"running {alg_name} on {len(data)} traces.")
        if fit:
            tsne_obj = algorithm.fit_transform(data)
        else:
            tsne_obj = algorithm.transform(data)
        # print(tsne_obj[0:5])

        # All the following complex stuff is for adding a 'show label on mouse over' feature
        # to the visualisation scatter graph.
        # It works when run from command line, but not in Jupyter/Spyder!
        # Surely there must be an easier way than doing all this...
        # Code adapted from:
        # https://stackoverflow.com/questions/55891285/how-to-make-labels-appear-
        #     when-hovering-over-a-point-in-multiple-axis/55892690#55892690
        fig, ax = plt.subplots()  # figsize=(8, 6))  # 25% larger, for better printing
        if xlim:
            ax.set_xlim(xlim)
        if ylim:
            ax.set_ylim(ylim)
        if cmap is None:
            # Choose a default colormap.  See bottom of the matplotlib page:
            #   https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
            cmap = pltcm.get_cmap('brg')  # sequential map with nice b&w printing.
        elif isinstance(cmap, str):
            cmap = pltcm.get_cmap(cmap)  # it is the name of a matplotlib color map
        if markers is None:
            markers = "o"
        if isinstance(markers, str) and len(markers) > 1:
            # loop through the marker styles
            clusters = np.ma.array(self._clusters)
            markchars = markers + "o" * num_clusters
            for curr in range(max(num_clusters, len(markers))):
                #prepare for masking arrays - 'conventional' arrays won't do it
                mask = clusters != curr  # False means unmasked
                x_masked = np.ma.array(tsne_obj[:, 0], mask=mask)
                y_masked = np.ma.array(tsne_obj[:, 1], mask=mask)
                color = cmap(curr / num_clusters)
                # c_masked = np.ma.array(clusters, mask=mask)
                # print(f"DEBUG:  mark {curr} is '{markers[curr]}' x={x_masked[0:10]} cl={c_masked[0:10]} color={color}")
                sc = ax.plot(x_masked, y_masked, color=color, linewidth=0,
                             label=f"c{curr}",
                             marker=markchars[curr],
                             markersize=markersize)
            leg = ax.legend(loc='best') #, ncol=2, mode="expand", shadow=True, fancybox=True)
            leg.get_frame().set_alpha(0.5)
        else:
            sc = plt.scatter(tsne_obj[:, 0], tsne_obj[:, 1], c=self._clusters,
                             cmap=cmap, marker=markers)

        if filename:
            plt.savefig(filename)
        names = [str(tr) for tr in self.traces]  # these are in same order as tsne_df rows.

        annot = ax.annotate("",
                            xy=(0, 0),
                            xytext=(0, 20),
                            textcoords="offset points",
                            bbox=dict(boxstyle="round", fc="w"),
                            arrowprops=dict(arrowstyle="->"),
                            )
        annot.set_visible(False)

        def update_annot(ind):
            pos = sc.get_offsets()[ind["ind"][0]]
            annot.xy = pos
            # text = "{}, {}".format(" ".join(list(map(str, ind["ind"]))),
            #                        " ".join([str(names[n]) for n in ind["ind"]]))
            anns = [f"{n} ({self._clusters[n]}): {str(names[n])}" for n in ind["ind"]]
            text = "\n".join(anns)
            annot.set_text(text)
            # annot.get_bbox_patch().set_facecolor(cmap(norm(c[ind["ind"][0]])))
            # annot.get_bbox_patch().set_alpha(0.4)

        def hover(event):
            vis = annot.get_visible()
            if event.inaxes == ax:
                cont, ind = sc.contains(event)
                if cont:
                    update_annot(ind)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if vis:
                        annot.set_visible(False)
                        fig.canvas.draw_idle()

        fig.canvas.mpl_connect("motion_notify_event", hover)
        plt.show(block=block)

    def get_cluster(self, num: int) -> List[Trace]:
        """Gets a list of all the Trace objects that are in the given cluster."""
        if self._clusters is None:
            raise Exception("You must call create_clusters() before get_cluster(_)")
        if len(self._clusters) != len(self.traces):
            raise Exception("Traces have changed, so you must call create_clusters() again.")
        return [tr for (i, tr) in zip(self._clusters, self.traces) if i == num]


class TraceEncoder(json.JSONEncoder):
    """Custom JSON encoder because objects from zeep could not be serialised.

    Based on ideas from this blog entry by 'The Fellow' (Ouma Rodgers):
    https://medium.com/python-pandemonium/json-the-python-way-91aac95d4041.

    This does not handle XML objects, as they should be decoded via xml_decode first.
    """

    def default(self, obj):
        if isinstance(obj, (dict, list, tuple, str, int, float, bool)):
            return super().default(obj)  # JSON already handles these
        if isinstance(obj, decimal.Decimal):
            return float(round(obj, 6))  # f"{o:.5f}"
        if isinstance(obj, (bytes, bytearray)):
            return "BYTES..."    # TODO: handle these better: repr(o)?
        if isinstance(obj, (set, frozenset)):
            return list(obj)
        if isinstance(obj, (datetime.date, datetime.datetime, datetime.time)):
            return obj.isoformat()  # as a string
        if hasattr(obj, "__dict__"):
            result = {
                "__class__": obj.__class__.__name__,
                "__module__": obj.__module__
                }
            if len(obj.__dict__) == 1 and "__values__" in obj.__dict__:
                # zeep seems to hide the attributes in a __values__ dict.
                # We lift them up to the top level to make the json more readable.
                self._add_public_attributes(result, obj.__dict__["__values__"])
            else:
                self._add_public_attributes(result, obj.__dict__)
            return result
        raise Exception("JSON serialisation not implemented yet for: " +
                        str(obj) + " type " + str(type(obj)) + " dir:" + ",".join(dir(obj)))

    def _add_public_attributes(self, result, attrs) -> None:
        for (name, value) in attrs.items():
            if not name.startswith("_"):
                result[name] = value


def xml_decode(obj: ET.Element) -> Union[str, Dict[str, Any]]:
    """Custom XML encoder to decode XML into a Python dictionary suitable for JSON encoding.

    This roughly follows the ideas from:
    https://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html.

    For simple XML objects with no attributes and no children, this returns just the text string.
    For more complex XML objects, it returns a dictionary.

    Note that the top-level tag of 'obj' is assumed to be handled by the caller.
    That is, the caller will typically do ```d[tag] = xml_decode(obj)``` where xml_decode
    will return either a simple string, or a dictionary.
    """
    if len(obj) == 0 and len(obj.attrib) == 0:
        return cast(str, obj.text)
    else:
        # return obj as a dictionary
        result: Dict[str, Any] = {}
        for (n, v) in obj.attrib.items():
            result[n] = v
        # child objects are more tricky, since some tags may appear multiple times.
        # If a tag appears multiple times, we map it to a list of child objects.
        curr_tag = ""
        curr_list: List[Union[str, Dict[str, Any]]] = []
        for child in obj:
            if child.tag != curr_tag:
                # save the child(ren) we have just finished
                if len(curr_list) > 0:
                    result[curr_tag] = curr_list if len(curr_list) > 1 else curr_list[0]
                curr_list = []
                curr_tag = child.tag
            curr_list.append(xml_decode(child))
        if len(curr_list) > 0:
            result[curr_tag] = curr_list if len(curr_list) > 1 else curr_list[0]
        if obj.text and obj.text.strip():  # ignore text that is just whitespace
            result["text"] = obj.text
        return result


def default_map_to_chars(actions: Set[str], given: Dict[str, str] = None) -> Dict[str, str]:
    """Tries to guess a useful default mapping from action names to single characters.

    Args:
        actions: the names of all the actions.
        given: optional pre-allocation of a few action names to chars.
            You can use this to override the default behaviour.

    Returns:
        A map from every name in actions to a unique single character.
    """
    names: List[str] = sorted(list(actions))
    result: Dict[str, str] = {} if given is None else given.copy()
    # TODO: a better algorithm might be to break up compound words and look for word prefixes?
    curr_prefix = ""
    pass2 = []
    for i in range(len(names)):
        name = names[i]
        if name in result:
            continue  # given
        # skip over any prefix that was in common with previous name.
        if name.startswith(curr_prefix):
            pos = len(curr_prefix)
        else:
            pos = 0
        # check ahead for common prefixes first
        if i + 1 < len(names):
            nxt = names[i + 1]
            if nxt.startswith(name) and name[0] not in result.values():
                result[name] = name[0]
                curr_prefix = name
                continue
            prefix = max([p for p in range(max(len(name), len(nxt))) if name[0:p] == nxt[0:p]])
            # print(f"  found prefix {prefix} of {name} and {nxt}")
            curr_prefix = name[0:prefix]
        else:
            prefix = 0
            curr_prefix = ""
        if prefix > 0 and prefix > pos:
            pos = prefix
        done = False
        for j in range(pos, len(name)):
            if name[pos] not in result.values():
                result[name] = name[pos]
                done = True
                break
        if not done:
            pass2.append(name)
    # Pass 2 (all visible ASCII chars except " and ')
    allchars = "".join([chr(n) for n in range(42, 127)]) + "!#$%&()"
    for name in pass2:
        for ch in name + allchars:
            if ch not in result.values():
                result[name] = ch
                break  # move onto next name in pass2
    return result


def all_action_names(traces: List[Trace]) -> Set[str]:
    """Collects all the action names that appear in the given traces."""
    result = set()
    for tr in traces:
        for ev in tr.events:
            action = ev.action
            result.add(action)
    return result


def trace_to_string(trace: List[Event], to_char: Mapping[str, str], compress: List[str] = None,
                    color_status: bool = False) -> str:
    """Converts a trace to a short summary string, one character per action.

    Args:
        trace: the sequence of JSON-like events, with an "action" field.
        to_char: maps each action name to a single character.  This map must include every
            action name that appears in the traces.  A suitable map can be constructed via
            TraceSet.get_event_chars().
        compress: a list of Action names.  Repeated events will be compressed if in this list.
        color_status: True means color the string red where status is non-zero.
            This uses ANSI escape sequences, so needs to be printed to a terminal.

    Returns:
        a summary string.
    """
    compress_set = set() if compress is None else set(compress)
    chars = []
    prev_action = None
    for ev in trace:
        action = ev.action
        if action == prev_action and action in compress_set:
            # NOTE: we color compressed output just based on the first event.
            pass
        else:
            if color_status and ev.status != 0:
                chars.append("\033[91m")  # start RED
                chars.append(to_char[action])
                chars.append("\033[0m")  # turn off color
            else:
                chars.append(to_char[action])
            prev_action = action
    return "".join(chars)


def traces_to_pandas(traces: List[Trace]) -> pd.DataFrame:
    """Collects all events into a single Pandas DataFrame.

    Columns include the trace number, the event number, the action name, each input parameter,
    the result status and error message.

    TODO: we could convert complex values to strings before sending to Pandas?
    TODO: we could have an option to encode strings into integer properties?
    """
    rows = []
    for tr_num in range(len(traces)):
        events = traces[tr_num].events
        for ev_num in range(len(events)):
            event = events[ev_num]
            row = {"Trace": tr_num, "Event": ev_num, "Action": event.action}
            # we add "Status" and "Error" first, so that those columns come before inputs.
            row["Status"] = event.status
            row["Error"] = event.error_message
            row.update(event.inputs.items())
            rows.append(row)
    return pd.DataFrame(rows)



# -*- coding: utf-8 -*-
"""
Analyse results of a testing run.

It prints a one-line summary of each trace, plus some general statistics.
Optionally, it can also report how many rows were added to each database table.

@author: utting@usc.edu.au
"""

import pandas as pd
import argparse
from pathlib import Path
import textwrap

import agilkia


def read_database_changes(before_csv: str, after_csv: str) -> pd.DataFrame:
    """Reads two files of database row counts and calculates tuples added to each table.

    Args:
        before_csv: name of CSV file containing the 'before' counts.
        after_csv: name of CSV file containing the 'after' counts.

    Returns:
        A Pandas table with an 'added' column for how many rows were added to each table.
    """
    before = pd.read_csv(before_csv)
    after = pd.read_csv(after_csv)
    col_msg = "ERROR: {} must have columns 'name', 'row_count', ..."
    if list(before.columns[0:2]) != ["name", "row_count"]:
        print(col_msg.format(before_csv))
    if list(after.columns[0:2]) != ["name", "row_count"]:
        print(col_msg.format(before_csv))
    # we use inner join to get the intersection of the two sets of tables.
    changes = pd.merge(before, after, how="inner", on="name", suffixes=("_before", "_after"))
    changes["added"] = changes["row_count_after"] - changes["row_count_before"]
    return changes


def make_action_status_table(df: pd.DataFrame) -> pd.DataFrame:
    """From TraceSet DataFrame, creates a table of Actions showing how many got Ok vs Error."""
    ok = df[df.Status == 0].groupby("Action").size()
    err = df[df.Status != 0].groupby("Action").size()
    data = pd.DataFrame({"Ok": ok, "Err": err})
    data.fillna(0, inplace=True, downcast="infer")
    data["Total"] = data.Ok + data.Err
    totals = data.sum().rename("Total")
    # add Totals row at bottom
    data = data.append(totals)
    # total = df.shape[0]  # number of rows = total event count
    # percents = (totals * 100.0 / total).rename("Percent")
    # data = data.append(percents)
    return data


def main():
    """A command line program that gives an overview of a set of generated traces."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-a", "--after", help="database row counts AFTER testing (*.csv)")
    parser.add_argument("-b", "--before", help="database row counts BEFORE testing (*.csv)")
    parser.add_argument("-c", "--chars", help="name of action-to-CHAR mapping file (*.csv)")
    parser.add_argument("-r", "--repeats", help="remove REPEATS of this action")
    parser.add_argument("-s", "--status", help="show STATUS in color (red=error)",
                        action="store_true")
    parser.add_argument("traces", help="traces file (*.json)")
    args = parser.parse_args()
    # print(f"Args are:", args)
    if args.before and args.after:
        changes = read_database_changes(args.before, args.after)
        nonzero = changes[changes.added > 0].sort_values(by="added", ascending=False)
        print("==== database changes ====")
        print(nonzero)
    traceset = agilkia.TraceSet.load_from_json(Path(args.traces))
    actions = agilkia.all_action_names(traceset.traces)
    if args.chars:
        mapfile = pd.read_csv(args.chars, header=None)
        # we assume this has just two columns: 0=action_name and 1=char.
        char_map = dict(zip(mapfile.iloc[:, 0], mapfile.iloc[:, 1]))
        # print("given map=", char_map)
        traceset.set_event_chars(char_map)
    # print("final map=", char_map)
    repeats = [] if args.repeats is None else [args.repeats]
    for tr in traceset.traces:
        print(tr.to_string(compress=repeats, color_status=args.status))
    print("==== statistics ====")
    df = traceset.to_pandas()
    # print(df.head())
    print(f"Number of traces     : {len(traceset.traces)}")
    print(f"Average trace length : {df.groupby('Trace').count().Action.mean()}")
    print(f"Number of events     : {df.shape[0]}")
    print(f"Number of event kinds: {len(actions)}")
    print(textwrap.indent(str(make_action_status_table(df)), "    "))
    statuses = df.Status.value_counts()
    percent_ok = 100.0 * statuses[0] / df.shape[0]
    print(f"Detailed error counts: ({100.0 - percent_ok:.2f}%)")
    print(textwrap.indent(str(df.groupby("Error").count().Action), "    "))
    print(f"Percent of status ok : {percent_ok:.2f}%")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Unit tests for the RandomTester class.

TODO:
    * test the examples/scanner scripts

@author: utting@usc.edu.au
"""

import unittest
import random
from pathlib import Path
import subprocess

import agilkia

THIS_DIR = Path(__file__).parent
SCANNER_DIR = THIS_DIR.parent / "examples" / "scanner"


class TestExamplesScanner(unittest.TestCase):

    def test_generate(self):
        status = subprocess.check_call(["python", "generate_missing_tests_scanette.py", "--test"],
                                  cwd=SCANNER_DIR)
        with open("test_generate.log", "w") as out:
            out.write("status=" + str(status))

    def test_analyze(self):
        status = subprocess.check_call(["python", "analyse_scanette2.py", "--test"],
                                  cwd=SCANNER_DIR)
        with open("test_analyse.log", "w") as out:
            out.write("status=" + str(status))



"""Automated smart testing strategies for web services.

This 'agilkia' package is for testing web services and managing set of traces.
Traces may come from user interactions, or from automated test suites, etc.

The main data structure for traces is the ``TraceSet``:
* class TraceSet supports loading/saving traces as JSON, converting to Pandas, etc.
* class Trace is used by TraceSet, and contains a list of Events.
* Each Event is a dict that contains at least the following keys:
  - "action" gives the name of the action (a string);
  - "inputs" is a dict of input parameter names to values;
  - "outputs" is a dict of output parameter names to values.

Automated test generation facilities include:
* RandomTester generates random test sequences.
* SmartTester generates tests from an ML model
  (Currently this is included in RandomTester.generate_trace_ml,
  but this will be split into a separate class shortly).
"""

# This package follows a 'Convenience Store' model.
# That is, it directly exports all the features that will be useful to users.
# They do not need to import sub-modules.
#
# See the article: "What’s __init__ for me?" by Jacob Deppen on TowardsDataScience.com:
# https://towardsdatascience.com/whats-init-for-me-d70a312da583

__version__ = '0.5.1'

from . random_tester import (read_input_rules, uniq, build_interface, print_signatures,
                             TracePrefixExtractor, RandomTester, SmartSequenceGenerator,
                             DUMP_WSDL, DUMP_SIGNATURES, GOOD_PASSWORD, TRACE_END)
from . json_traces import (Event, Trace, TraceSet, TraceEncoder, TRACE_SET_VERSION,
                           MetaData, xml_decode, all_action_names, safe_name,
                           default_map_to_chars, trace_to_string, traces_to_pandas)



# -*- coding: utf-8 -*-
"""
Generate and execute tests for a Web Service.

Example args: http://www.soapclient.com/xml/soapresponder.wsdl

@author: utting@usc.edu.au
"""

import random
import pandas as pd        # type: ignore
import argparse
from pathlib import Path
from joblib import load    # type: ignore

import agilkia


def main():
    """A command line program to generate and execute test traces for a SOAP web service."""
    parser = argparse.ArgumentParser(description=__doc__)
    # parser.add_argument("-c", "--compress", help="compress repeats of this action")
    parser.add_argument("-n", "--tests", type=int, default=2, help="number of TESTS to generate")
    parser.add_argument("-l", "--length", type=int, default=5, help="LENGTH of each test")
    fmt = "(CSV with Name,Frequency,Value columns)"
    parser.add_argument("-i", "--inputs", help=f"INPUT values file: {fmt}")
    parser.add_argument("-c", "--chars", help="name of action-to-CHAR mapping file (*.csv)")
    parser.add_argument("-m", "--methods", help="METHODS to test (comma-separated)")
    parser.add_argument("--model", help="ML MODEL (*.joblib) to predict next action.")
    parser.add_argument("-v", "--verbose",
                        help="print VERBOSE messages during testing", action="store_true")
    parser.add_argument("-s", "--seed", type=int, help="SEED for random generator")
    parser.add_argument("-o", "--output", type=str, default="out.json", help="name of OUTPUT file")
    parser.add_argument("url", nargs='+', help="URL of web service server")
    args = parser.parse_args()
    # print(f"Args are:", args)

    json_output = Path(args.output).with_suffix(".json")
    if json_output.exists():
        raise Exception(f"do not want to overwrite {json_output}.")
    # process each optional argument
    action_chars = None
    if args.chars:
        mapfile = pd.read_csv(args.chars, header=None)
        # we assume this has just two columns: 0=action_name and 1=char.
        action_chars = dict(zip(mapfile.iloc[:, 0], mapfile.iloc[:, 1]))
    input_rules = None
    if args.inputs:
        input_rules = agilkia.read_input_rules(Path(args.inputs))
    methods = None
    if args.methods:
        methods = args.methods.split(",")
    rand = random.Random(args.seed)  # seed can be None
    model = None
    if args.model:
        model = load(args.model)
    print(f"Starting to test {args.url}")
    tester = agilkia.RandomTester(args.url, rand=rand, verbose=args.verbose,
                                  action_chars=action_chars, input_rules=input_rules,
                                  methods_to_test=methods)
    # get a password if needed
    if input_rules is not None and "username" in input_rules:
        # this will prompt for a password
        tester.set_username(input_rules["username"])
    # TODO: methods_to_test=None,
    for i in range(args.tests):
        if args.model:
            trace = tester.generate_trace_ml(model, length=args.length, start=True)
        else:
            trace = tester.generate_trace(length=args.length, start=True)
        if args.verbose:
            print(f"  {str(trace)}")
    tester.trace_set.save_to_json(json_output)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Unit tests for the RandomTester class.

TODO:
    * test the ML test generator

@author: utting@usc.edu.au
"""

import unittest
import random
from pathlib import Path
import sklearn.utils.estimator_checks

import agilkia

THIS_DIR = Path(__file__).parent
WSDL_EG = "http://www.soapclient.com/xml/soapresponder.wsdl"

test_input_rules = {
    "username": ["User1"],
    "password": ["<GOOD_PASSWORD>"] * 9 + ["bad-pass"],
    "speed": [str(s) for s in range(0, 120, 10)],
    "bstrParam1": ["VAL1"],
    "bstrParam2": ["p2AAA", "p2BBB"],
}


class TestReadInputRules(unittest.TestCase):

    def test_1(self):
        rules = agilkia.read_input_rules(THIS_DIR / "fixtures/inputs1.csv")
        self.assertEqual(["one"], rules["bstrParam1"])
        self.assertEqual(['two', 'two', 'two', 'TWO!'], rules["bstrParam2"])


class TestRandomTester(unittest.TestCase):

    def setUp(self):
        self.tester = agilkia.RandomTester(
                WSDL_EG,
                input_rules=test_input_rules,
                rand=random.Random(1234))

    def test_input_user(self):
        self.assertEqual("User1", self.tester.choose_input_value("username"))

    def test_input_password(self):
        self.assertEqual(agilkia.GOOD_PASSWORD, self.tester.choose_input_value("password"))

    def test_input_speeds(self):
        speeds = set()
        for i in range(100):
            speeds.add(self.tester.choose_input_value("speed"))
        self.assertEqual(12, len(speeds))  # all results should be covered

    def test_signature(self):
        sig = self.tester.get_methods()
        self.assertEqual(1, len(sig))
        self.assertEqual({"Method1"}, sig.keys())
        msig = sig["Method1"]
        self.assertEqual(1, len(msig))
        self.assertEqual({"input"}, msig.keys())
        self.assertEqual({"bstrParam1", "bstrParam2"}, msig["input"].keys())
        param1_details = "{'optional': False, 'type': 'String(value)'}"
        self.assertEqual(param1_details, str(msig["input"]["bstrParam1"]))

    def test_dummy_client_meta(self):
        """Test the dummy web service provided by soapresponder."""
        tester = agilkia.RandomTester(WSDL_EG,
                                      input_rules=test_input_rules,
                                      rand=random.Random(1234))
        meta_keys = ["date", "author", "dataset", "source",
                     "web_services", "methods_to_test", "input_rules",
                     "method_signatures"]
        mdata = tester.trace_set.meta_data
        for k in meta_keys:
            self.assertTrue(k in mdata, msg=k + " expected in meta_data")
        self.assertEqual(f"RandomTester", mdata["source"])
        self.assertEqual([WSDL_EG], mdata["web_services"])
        # check the signature
        self.assertEqual(set(["Method1"]), set(mdata["method_signatures"].keys()))
        sig = {'input': {
               'bstrParam1': {'optional': False, 'type': 'String(value)'},
               'bstrParam2': {'optional': False, 'type': 'String(value)'}}}
        self.assertEqual(sig, mdata["method_signatures"]["Method1"])

    def test_dummy_client0(self):
        """Test the dummy web service provided by soapresponder."""
        tester = agilkia.RandomTester(WSDL_EG, verbose=True,
                                      input_rules=test_input_rules,
                                      rand=random.Random(1234))
        print("Methods:", tester.get_methods())
        out1 = tester.call_method("Method1")
        expect = {"Status": 0, "value": "Your input parameters are VAL1 and p2AAA"}
        self.assertEqual(expect, out1.outputs)
        out1 = tester.call_method("Method1")
        self.assertEqual(expect, out1.outputs)
        out1 = tester.call_method("Method1")
        self.assertEqual(expect, out1.outputs)
        out1 = tester.call_method("Method1")
        expect["value"] = "Your input parameters are VAL1 and p2BBB"
        self.assertEqual(expect, out1.outputs)
        self.assertEqual(4, len(tester.curr_events))
        self.assertEqual(1, len(tester.trace_set.traces))
        # now generate a second trace
        tester.generate_trace(start=True, length=3)
        self.assertEqual(3, len(tester.curr_events))
        self.assertEqual(2, len(tester.trace_set.traces))
        # now test saving and loading those traces.
        traceset1 = tester.trace_set
        tmp_json = Path("tmp_dummy1.json")
        traceset1.save_to_json(tmp_json)
        traceset2 = agilkia.TraceSet.load_from_json(tmp_json)
        self.assertEqual(traceset2.meta_data, traceset1.meta_data)
        self.assertEqual(len(traceset2.traces), len(traceset1.traces))
        self.assertEqual(traceset2.traces[0].events[0].action,
                         traceset1.traces[0].events[0].action)
        tmp_json.unlink()

    def test_generate_trace(self):
        tr = self.tester.generate_trace()
        self.assertTrue(isinstance(tr, agilkia.Trace))
        self.assertEqual(20, len(tr.events))

    def test_decode_outputs(self):
        self.assertEqual({'Status': 0, "value": "abc"}, self.tester.decode_outputs("abc"))
        self.assertEqual({'Status': 0, "a": 2}, self.tester.decode_outputs({"a": 2}))
        # Also, zeep XML object outputs are tested in test_dummy_client0 above.


class TestTracePrefixExtractor(unittest.TestCase):

    ev1 = agilkia.Event("Order", {"Name": "Mark"}, {"Status": 0})
    ev2 = agilkia.Event("Skip", {"Size": 3}, {"Status": 1, "Error": "Too big"})
    ev3 = agilkia.Event("Pay", {"Name": "Mark", "Amount": 23.45}, {"Status": 0})
    
    def test_bag_of_words(self):
        tr1 = agilkia.Trace([self.ev1, self.ev2])
        tr2 = agilkia.Trace([self.ev3])
        traces = agilkia.TraceSet([tr1, tr1, tr2])
        self.assertEqual(3, len(traces))
        sut = agilkia.TracePrefixExtractor()
        sut.fit(traces)
        self.assertEqual(["Order", "Pay", "Skip"], sut.get_feature_names())
        X = sut.transform(traces)
        y = sut.get_labels()
        self.assertEqual((8, 3), X.shape)
        self.assertEqual(8, len(y))
        for row in [0, 3, 6]:
            self.assertEqual([0.0, 0.0, 0.0], X.iloc[row, :].tolist())
            self.assertEqual("Order" if row < 6 else "Pay", y[row])
        for row in [2, 5]:
            self.assertEqual([1.0, 0.0, 1.0], X.iloc[row, :].tolist())
            self.assertEqual(agilkia.TRACE_END, y[row])
        self.assertEqual([0.0, 1.0, 0.0], X.iloc[7, :].tolist())
        
        


#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['agilkia']

package_data = \
{'': ['*']}

install_requires = \
['flit>=1.3',
 'ipython>=7.8.0',
 'jsonpickle>=1.2',
 'matplotlib>=3.1.1',
 'numpy>=1.17.2',
 'pandas>=0.25.1',
 'pillow>=6.1.0',
 'pip>=19.2.2',
 'requests>=2.22.0',
 'scikit-learn>=0.21.2',
 'scipy>=1.3.1',
 'zeep>=3.4.0']

extras_require = \
{'test': ['pytest >=5.1.0', 'pytest-cov>=2.7.1', 'mypy>=0.740']}

setup(name='agilkia',
      version='0.5.1',
      description='Automated smart testing strategies for web services.',
      author='Mark Utting',
      author_email='utting@usc.edu.au',
      url='https://github.com/utting/bus-testing',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7,<3.8',
     )



# -*- coding: utf-8 -*-
"""
Test JSON saving and loading.

@author: utting@usc.edu.au
"""

import agilkia
import jsonpickle     # type: ignore
import json
import decimal
import datetime
import xml.etree.ElementTree as ET
from pathlib import Path
import pandas as pd   # type: ignore
import numpy.testing as nptest
import unittest
import pytest         # type: ignore

THIS_DIR = Path(__file__).parent


class Dummy():
    """Dummy object for testing JSON saving/loading of custom objects."""
    def __init__(self):
        self.f = [3.14]


class TestTraceEncoder(unittest.TestCase):
    """Unit Tests for agilkia.TraceEncoder."""

    xml0 = '{"__class__": "Element", "__module__": "xml.etree.ElementTree", "__tag__": '
    xml1 = xml0 + '"Inner", "__text__": null, "__children__": [], "first": 1, "second": 22}'
    xml2 = xml0 + '"Outer", "__text__": null, "__children__": [' + xml1 + '], "size": 123}'

    def dumps(self, obj):
        """Convenience function that calls json.dumps with appropriate arguments."""
        return json.dumps(obj, cls=agilkia.TraceEncoder)

    def test_decimal(self):
        d1 = decimal.Decimal(3.45)
        d2 = decimal.Decimal(3.4500048012)
        self.assertEqual('3.45', self.dumps(d1))
        self.assertEqual('3.450005', self.dumps(d2))

    def test_object(self):
        d1 = Dummy()
        str1 = '{"__class__": "Dummy", "__module__": "' + __name__ + '", "f": [3.14]}'
        self.assertEqual(str1, self.dumps(d1))

    def test_nested_object(self):
        d1 = Dummy()
        d2 = Dummy()
        d2.extra = d1
        str1 = '{"__class__": "Dummy", "__module__": "' + __name__ + '", "f": [3.14]}'
        str2 = str1[0:-1] + ', "extra": ' + str1 + '}'
        self.assertEqual(str2, self.dumps(d2))

    def test_dict(self):
        d1 = {"a": 1, "b": [2, 3]}
        str1 = '{"a": 1, "b": [2, 3]}'
        self.assertEqual(str1, self.dumps(d1))

    def test_set(self):
        d = {'testing': {1, 2, 3}}
        str1 = '{"testing": [1, 2, 3]}'
        self.assertEqual(str1, self.dumps(d))

    def test_datetime(self):
        d = datetime.datetime(2019, 9, 17, hour=18, minute=58)
        self.assertEqual('"2019-09-17T18:58:00"', self.dumps(d))

    def test_time(self):
        t = datetime.time(hour=18, minute=58)
        self.assertEqual('"18:58:00"', self.dumps(t))

    @pytest.mark.skip(reason="XML objects should be handled separately, by top level.")
    def test_xml1(self):
        xml1 = ET.Element("Inner", attrib={"first": 1, "second": 22})
        self.assertEqual(self.xml1, self.dumps(xml1))
        xml2 = ET.Element("Outer", size=123)
        xml2.append(xml1)
        self.assertEqual(self.xml2, self.dumps(xml2))

    def test_trace(self):
        ev1 = agilkia.Event("Order", {"Name": "Mark"}, {"Status": 0})
        tr1 = agilkia.Trace([ev1])
        s0 = '{"__class__": "Trace", "__module__": "agilkia.json_traces", "events": ['
        s1a = '{"__class__": "Event", "__module__": "agilkia.json_traces", "action": "Order", '
        s1b = '"inputs": {"Name": "Mark"}, "outputs": {"Status": 0}, "meta_data": {}}'
        s2 = '], "meta_data": {}}'
        expect = s0 + s1a + s1b + s2
        self.maxDiff = None
        self.assertEqual(expect, self.dumps(tr1))


class TestXMLDecode(unittest.TestCase):
    """Unit Tests for agilkia.xml_decode."""

    inner = {'first': 1, 'second': 22}

    def test_simple(self):
        xml0 = ET.Element("Simple")
        xml0.text = "abc "
        self.assertEqual("abc ", agilkia.xml_decode(xml0))

    def test_attributes(self):
        xml1 = ET.Element("Inner", attrib=self.inner)
        self.assertEqual(self.inner, agilkia.xml_decode(xml1))

    def test_empty_text(self):
        xml1 = ET.Element("Inner", size=1234)
        xml1.text = "\n    "
        self.assertEqual({'size': 1234}, agilkia.xml_decode(xml1))

    def test_children(self):
        xml = ET.Element("Outer", size=3)
        xml.append(ET.Element("Inner", attrib={"first": 1, "second": 22}))
        vals = ["abc", "def", ""]
        for s in vals:
            x = ET.Element("Child")
            x.text = s
            xml.append(x)
        self.assertEqual({'size': 3, 'Inner': self.inner, 'Child': vals}, agilkia.xml_decode(xml))


class TestJsonTraces(unittest.TestCase):
    """Tests for loading and saving Event, Trace and TraceSet objects."""

    def test_round_trip(self):
        """Test that load and save are the inverse of each other."""
        data2 = agilkia.TraceSet.load_from_json(THIS_DIR / "fixtures/traces1.json")
        self.assertEqual(agilkia.TRACE_SET_VERSION, data2.version)
        tmp2_json = Path("tmp2.json")
        data2.save_to_json(tmp2_json)
        data3 = agilkia.TraceSet.load_from_json(tmp2_json)
        self.assertEqual(agilkia.TRACE_SET_VERSION, data3.version)
        self.assertEqual(data2.meta_data, data3.meta_data)
        assert len(data2.traces) == len(data3.traces)
        for i in range(len(data2.traces)):
            # we have not defined equality on Trace objects, so just compare first events.
            ev3 = data3.traces[i].events[0]
            ev2 = data2.traces[i].events[0]
            self.assertEqual(ev3.action, ev2.action)
            self.assertEqual(ev3.inputs, ev2.inputs)
            self.assertEqual(ev3.status, ev2.status)
        # if all went well, we can delete the temp file now.
        tmp2_json.unlink()

    def test_pickled_round_trip(self):
        """Loads some pickled zeep objects and checks that they save/load okay."""
        pickled = THIS_DIR / "fixtures/traces_pickled.json"
        # with open(traces_file, "r") as input:
        data = jsonpickle.loads(pickled.read_text())
        print(len(data), "traces loaded")
        parent = agilkia.TraceSet([], {"date": "2019-10-02", "dataset": "test1"})
        for events in data:
            trace = agilkia.Trace([])
            for e in events:
                event = agilkia.Event(e["action"], e["inputs"], e["outputs"])
                trace.append(event)
            parent.append(trace)
        tmp_json = Path("tmp.json")
        tmp2_json = Path("tmp2.json")
        parent.save_to_json(tmp_json)
        parent2 = agilkia.TraceSet.load_from_json(tmp_json)
        assert len(data) == len(parent2.traces)

        parent2.save_to_json(tmp2_json)
        parent3 = agilkia.TraceSet.load_from_json(tmp2_json)
        assert len(data) == len(parent3.traces)
        for i in range(len(parent3.traces)):
            ev3 = parent3.traces[i].events[0]
            ev2 = parent2.traces[i].events[0]
            self.assertEqual(ev3.action, ev2.action)
            self.assertEqual(ev3.inputs, ev2.inputs)
        # if all went well, we can delete the temp files now.
        tmp_json.unlink()
        tmp2_json.unlink()


class TestEvent(unittest.TestCase):
    """Unit tests for agilkia.Event."""

    ev1 = agilkia.Event("Order", {"Name": "Mark"}, {"Status": -2, "Error": "unknown name"})
    ev2 = agilkia.Event("Order", {"Name": "Other"}, {})

    def test_status(self):
        self.assertEqual(-2, self.ev1.status)
        self.assertEqual(0, self.ev2.status)

    def test_error_message(self):
        self.assertEqual("unknown name", self.ev1.error_message)
        self.assertEqual("", self.ev2.error_message)

    def test_str(self):
        self.assertEqual("Event(Order, {'Name': 'Other'}, {})", str(self.ev2))


class TestTrace(unittest.TestCase):
    """Unit tests for agilkia.Trace and agilkia.TraceSet."""

    ev1 = agilkia.Event("Order", {"Name": "Mark"}, {"Status": 0})
    ev2 = agilkia.Event("Skip", {"Size": 3}, {"Status": 1, "Error": "Too big"})
    ev3 = agilkia.Event("Pay", {"Name": "Mark", "Amount": 23.45}, {"Status": 0})
    ev4 = agilkia.Event("Ski", {"Type": "downhill"}, {"Status": 1})
    to_char = {"Order": "O", "Skip": ",", "Pay": "p"}

    def test_trace(self):
        tr1 = agilkia.Trace([self.ev2, self.ev1, self.ev3])  # no parent initially
        with self.assertRaises(Exception):
            tr1.to_string()
        self.assertEqual("???", str(tr1))
        self.assertEqual(3, len(tr1))
        self.assertEqual(self.ev2, tr1[0])
        self.assertEqual(self.ev3, tr1[-1])

    def test_traceset(self):
        parent = agilkia.TraceSet([], {})
        # add a first trace
        tr1 = agilkia.Trace([self.ev2, self.ev1, self.ev3])
        parent.append(tr1)
        self.assertEqual(parent, tr1.trace_set())
        self.assertEqual("SOP", tr1.to_string())
        self.assertEqual("SOP", str(tr1))
        # now add a second trace.
        tr2 = agilkia.Trace([self.ev4, self.ev2])
        parent.append(tr2)
        self.assertEqual("Sp", tr2.to_string())
        self.assertEqual("Sp", str(tr2))
        self.assertEqual("pOP", str(tr1))  # changed since to-char is recalculated
        self.assertEqual(2, len(parent))
        self.assertEqual(tr1, parent[0])
        self.assertEqual(tr2, parent[-1])

    def test_trace_iter(self):
        tr1 = agilkia.Trace([self.ev2, self.ev1, self.ev3])
        it = iter(tr1)
        self.assertEqual(self.ev2, next(it))
        self.assertEqual(self.ev1, next(it))
        self.assertEqual(self.ev3, next(it))
        with self.assertRaises(StopIteration):
            next(it)

    def test_simple(self):
        tr1 = agilkia.Trace([self.ev1, self.ev2, self.ev3])
        s = tr1.to_string(to_char=self.to_char)
        self.assertEqual("O,p", s)

    def test_compress(self):
        events = [self.ev2, self.ev2, self.ev1, self.ev2, self.ev3, self.ev2, self.ev2, self.ev2]
        tr1 = agilkia.Trace(events)
        s = tr1.to_string(to_char=self.to_char)
        self.assertEqual(",,O,p,,,", s)
        s = tr1.to_string(to_char=self.to_char, compress=["Skip"])
        self.assertEqual(",O,p,", s)

    def test_status(self):
        tr = agilkia.Trace([self.ev1, self.ev2, self.ev3])
        s = tr.to_string(to_char=self.to_char, color_status=True)
        self.assertEqual("O\033[91m,\033[0mp", s)

    def test_all_action_names(self):
        tr1 = agilkia.Trace([self.ev1, self.ev3])
        tr2 = agilkia.Trace([self.ev2, self.ev3])
        self.assertEqual(set(["Order", "Skip", "Pay"]), agilkia.all_action_names([tr1, tr2]))

    def test_pandas(self):
        traces = agilkia.TraceSet([])
        traces.append(agilkia.Trace([self.ev1, self.ev3]))
        traces.append(agilkia.Trace([self.ev2, self.ev3]))
        df = traces.to_pandas()
        self.assertEqual(4, df.shape[0])  # rows
        self.assertEqual(8, df.shape[1])  # columns
        cols = ["Trace", "Event", "Action", "Status", "Error", "Name", "Amount", "Size"]
        self.assertEqual(cols, list(df.columns))

    def test_default_meta_data(self):
        now = str(datetime.datetime.now())
        md = agilkia.TraceSet.get_default_meta_data()
        self.assertEqual("pytest", md["source"].split("/")[-1])
        self.assertEqual(now[0:10], md["date"][0:10])  # same date, unless it is exactly midnight!

    def test_arff_type(self):
        traces = agilkia.TraceSet([])
        i64 = pd.api.types.pandas_dtype("int64")
        self.assertEqual("INTEGER", traces.arff_type(i64))

    def test_split_none(self):
        # each ev1 will start a new trace
        trace = agilkia.Trace([self.ev2, self.ev1, self.ev3, self.ev1, self.ev1, self.ev2])
        traces = agilkia.TraceSet([trace])
        with self.assertRaises(Exception):
            traces.with_traces_split()
        traces2 = traces.with_traces_split(start_action="XYZ")
        self.assertEqual(1, len(traces2))  # none split

    def test_split_action(self):
        # each ev1 will start a new trace
        trace = agilkia.Trace([self.ev2, self.ev1, self.ev3, self.ev1, self.ev1, self.ev2])
        traces = agilkia.TraceSet([trace])
        traces2 = traces.with_traces_split(start_action="Order")
        self.assertEqual(4, len(traces2))
        self.assertEqual("Skip", traces2[0].events[0].action)
        self.assertEqual("Order", traces2[1].events[0].action)
        self.assertEqual("Order", traces2[2].events[0].action)
        self.assertEqual("Order", traces2[3].events[0].action)

    def test_split_input(self):
        ev3b = agilkia.Event("Pay", {"Name": "Merry", "Amount": 23.45}, {"Status": 0})
        # each change in the "Name" input will start a new trace
        trace = agilkia.Trace([self.ev1, self.ev3, ev3b, self.ev3, self.ev2, self.ev1])
        traces = agilkia.TraceSet([trace])
        traces2 = traces.with_traces_split(input_name="Name")
        #for t in traces2:
        #    print(t)
        self.assertEqual(3, len(traces2))
        self.assertEqual("Mark", traces2[0].events[0].inputs["Name"])
        self.assertEqual("Merry", traces2[1].events[0].inputs["Name"])
        self.assertEqual("Mark", traces2[2].events[0].inputs["Name"])
        self.assertEqual(2, len(traces2[0]))
        self.assertEqual(1, len(traces2[1]))
        self.assertEqual(3, len(traces2[2]))

    def test_group_input(self):
        ev3b = agilkia.Event("Pay", {"Name": "Merry", "Amount": 23.45}, {"Status": 0})
        # each different "Name" input will be grouped into a new trace
        trace = agilkia.Trace([self.ev1, self.ev3, ev3b, self.ev3, self.ev2, self.ev1])
        traces = agilkia.TraceSet([trace])
        traces2 = traces.with_traces_grouped_by("Name")  # self.ev2 will be discarded
        self.assertEqual(2, len(traces2))
        self.assertEqual("Mark", traces2[0].events[0].inputs["Name"])
        self.assertEqual("Merry", traces2[1].events[0].inputs["Name"])
        self.assertEqual(4, len(traces2[0]))
        self.assertEqual(1, len(traces2[1]))


class TestTraceSet(unittest.TestCase):
    """Unit tests specifically for agilkia.TraceSet.

    TODO: move some of the above tests into this class too.
    """

    ev1 = agilkia.Event("Order", {"Name": "Mark"}, {"Status": 0})
    ev2 = agilkia.Event("Skip", {"Size": 3}, {"Status": 1, "Error": "Too big"})

    def test_meta_data_copy(self):
        tr1 = agilkia.Trace([self.ev1])
        tr2 = agilkia.Trace([self.ev2])
        traces1 = agilkia.TraceSet([tr1, tr2])
        traces1.set_meta("dataset", "Copy Test")
        traces2 = agilkia.TraceSet([tr1, tr2])
        self.assertEqual("Copy Test", traces2.get_meta("dataset"))
        # but now move tr2 into a different trace set.
        traces3 = agilkia.TraceSet([tr2])
        self.assertEqual("Copy Test", traces3.set_meta("dataset", "Different Parent"))
        traces4 = agilkia.TraceSet([tr1, tr2])  # different parents
        self.assertEqual("unknown", traces4.get_meta("dataset"))

    def test_clustering(self):
        tr1 = agilkia.Trace([self.ev2, self.ev2, self.ev1]) # in cluster 1
        tr2 = agilkia.Trace([self.ev1, self.ev1, self.ev2]) # in cluster 0
        tr3 = agilkia.Trace([self.ev1, self.ev2, self.ev1]) # in cluster 0
        traces1 = agilkia.TraceSet([tr1, tr2, tr3])
        data = traces1.get_trace_data()
        nptest.assert_array_equal(["Order", "Skip"], data.columns)
        self.assertFalse(traces1.is_clustered())
        n = traces1.create_clusters(data)
        self.assertEqual(2, n)
        self.assertTrue(traces1.is_clustered())
        nptest.assert_array_equal([tr2, tr3], traces1.get_cluster(0))
        nptest.assert_array_equal([tr1], traces1.get_cluster(1))
        #
        # Now test save then load - clusters should be lost, since they are currently transient.
        tmp3_json = Path("tmp3.json")
        traces1.save_to_json(tmp3_json)
        traces2 = agilkia.TraceSet.load_from_json(tmp3_json)
        self.assertFalse(traces2.is_clustered())
        tmp3_json.unlink()


class TestDefaultMapToChars(unittest.TestCase):

    def test_default_map_to_chars(self):
        actions = ["Order", "Skip", "PayLate", "Pay"]
        expect = {"Order": "O", "Skip": "S", "PayLate": "L", "Pay": "P"}
        self.assertEqual(expect, agilkia.default_map_to_chars(actions))

    def test_default_map_to_chars_prefixes(self):
        actions = ["Order", "PayLate", "PayEarly", "PayExtra"]
        expect = {'Order': 'O', 'PayEarly': 'a', 'PayExtra': 'x', 'PayLate': 'L'}
        self.assertEqual(expect, agilkia.default_map_to_chars(actions))

    def test_default_map_to_chars_hard(self):
        actions = ["O", "Oa", "Oy", "Pay", "Play", "yay"]
        expect = {'O': 'O', 'Oa': 'a', 'Oy': 'y', 'Pay': 'P', 'Play': 'l', 'yay': '*'}
        self.assertEqual(expect, agilkia.default_map_to_chars(actions))

    def test_default_map_to_chars_given(self):
        actions = ["Order", "Save", "Skip", "PayLate", "Pay"]
        given = {"Save": "."}
        expect = {'Order': 'O', 'Pay': 'P', 'PayLate': 'L', 'Save': '.', 'Skip': 'S'}
        self.assertEqual(expect, agilkia.default_map_to_chars(actions, given=given))


class TestSafeString(unittest.TestCase):

    def test_safe_string(self):
        self.assertEqual("Ab9", agilkia.safe_name("Ab9"))
        self.assertEqual("_Ab9_etc_json", agilkia.safe_name("!Ab9 etc.json"))


if __name__ == "__main__":
    unittest.main()
